from tala.model.common import Modality
from tala.model.speaker import Speaker
from tala.model.semantic_object import SemanticObject, OntologySpecificSemanticObject, SemanticObjectWithContent
from tala.utils.as_json import convert_to_json
from tala.utils.unicodify import unicodify
from tala.utils import float_comparison


class MoveException(Exception):
    pass


class Move(SemanticObject):
    QUIT = "quit"
    GREET = "greet"
    NO_MOVE = "no_move"
    MUTE = "mute"
    UNMUTE = "unmute"

    ASK = "ask"
    ANSWER = "answer"
    REQUEST = "request"
    REPORT = "report"
    PREREPORT = "prereport"

    def __init__(
        self,
        type,
        understanding_confidence=None,
        speaker=None,
        modality=None,
        weighted_understanding_confidence=None,
        utterance=None,
        ddd_name=None,
        perception_confidence=None
    ):
        SemanticObject.__init__(self)
        self._type = type
        self._perception_confidence = None
        self._understanding_confidence = None
        self._weighted_understanding_confidence = None
        self._speaker = None
        self._modality = None
        self._ddd_name = None
        self._background = None
        self._utterance = None
        if (
            understanding_confidence is not None or speaker is not None or modality is not None or utterance is not None
            or ddd_name is not None or perception_confidence is not None
        ):
            self.set_realization_data(
                understanding_confidence=understanding_confidence,
                speaker=speaker,
                modality=modality,
                utterance=utterance,
                ddd_name=ddd_name,
                perception_confidence=perception_confidence
            )
            if weighted_understanding_confidence is not None:
                self.weighted_understanding_confidence = weighted_understanding_confidence

    def is_move(self):
        return True

    def get_type(self):
        return self._type

    @property
    def understanding_confidence(self):
        return self._understanding_confidence

    @property
    def weighted_understanding_confidence(self):
        return self._weighted_understanding_confidence

    @weighted_understanding_confidence.setter
    def weighted_understanding_confidence(self, confidence):
        self._weighted_understanding_confidence = confidence

    @property
    def perception_confidence(self):
        return self._perception_confidence

    @property
    def confidence(self):
        confidence_sources = [self.perception_confidence, self.weighted_understanding_confidence]
        if None in confidence_sources:
            return None
        return self.perception_confidence * self.understanding_confidence

    @property
    def weighted_confidence(self):
        confidence_sources = [self.perception_confidence, self.weighted_understanding_confidence]
        if None in confidence_sources:
            return None
        return self.perception_confidence * self.weighted_understanding_confidence

    def get_speaker(self):
        return self._speaker

    def get_modality(self):
        return self._modality

    def get_utterance(self):
        return self._utterance

    def set_ddd_name(self, ddd_name):
        self._ddd_name = ddd_name

    def get_ddd_name(self):
        return self._ddd_name

    def set_realization_data(
        self,
        understanding_confidence=None,
        speaker=None,
        modality=None,
        utterance=None,
        ddd_name=None,
        perception_confidence=None
    ):
        self._verify_realization_data(
            understanding_confidence=understanding_confidence, speaker=speaker, modality=modality, ddd_name=ddd_name
        )
        self._perception_confidence = perception_confidence
        if self._perception_confidence is None:
            self._perception_confidence = 1.0
        self._understanding_confidence = understanding_confidence
        if self._understanding_confidence is None:
            self._understanding_confidence = 1.0
        self._weighted_understanding_confidence = self._understanding_confidence
        self._speaker = speaker
        self._modality = modality or Modality.SPEECH
        self._utterance = utterance
        self._ddd_name = ddd_name

    def _verify_realization_data(self, understanding_confidence=None, speaker=None, modality=None, ddd_name=None):
        if self._understanding_confidence is not None or self._speaker is not None or self._modality is not None or (
            self._speaker == Speaker.USR and self._ddd_name is not None
        ):
            raise MoveException("realization data already set")
        if speaker is None:
            raise MoveException("speaker must be supplied")
        if speaker == Speaker.SYS:
            if understanding_confidence is not None and understanding_confidence != 1.0:
                raise MoveException("understanding confidence below 1.0 not allowed for system moves")
        if speaker == Speaker.USR:
            if understanding_confidence is None:
                raise MoveException("understanding confidence must be supplied for user moves")
            if ddd_name is None:
                raise MoveException("ddd_name must be supplied")
        if modality not in Modality.MODALITIES:
            raise MoveException("unsupported modality: '%s'" % modality)

    def __hash__(self):
        return hash((
            self.__class__.__name__, self._type, self._perception_confidence, self._understanding_confidence,
            self._weighted_understanding_confidence, self._speaker, self._utterance, self._ddd_name
        ))

    def __unicode__(self):
        return self._get_expression(True)

    def __str__(self):
        return unicode(self).encode("utf-8")

    def __repr__(self):
        return str(self)

    def _get_expression(self, include_attributes):
        string = "Move("
        string += self._type
        if include_attributes:
            string += self._build_string_from_attributes()
        string += ")"
        return string

    def get_semantic_expression(self):
        return self._get_expression(False)

    def _build_string_from_attributes(self):
        string = ""
        if self._ddd_name:
            string += ", ddd_name=%r" % self._ddd_name
        if self._speaker:
            string += ", speaker=%s" % self._speaker
        if self._understanding_confidence is not None:
            string += ", understanding_confidence=%s" % self._understanding_confidence
        if self._weighted_understanding_confidence is not None:
            string += ", weighted_understanding_confidence=%s" % self._weighted_understanding_confidence
        if self.perception_confidence is not None:
            string += ", perception_confidence=%s" % self.perception_confidence
        if self._modality:
            string += ", modality=%s" % self._modality
        if self._utterance:
            string += ", utterance=%r" % self._utterance
        return string

    def __eq__(self, other):
        try:
            return (
                other.is_move() and self.move_content_equals(other) and self._speaker == other._speaker
                and self._are_understanding_confidences_equal(other) and self._are_perception_confidences_equal(other)
                and self._utterance == other._utterance and self._ddd_name == other._ddd_name
            )
        except AttributeError:
            return False

    @staticmethod
    def _is_confidence_equal(this, other):
        if this is None:
            return other is None
        return float_comparison.isclose(this, other)

    def _are_understanding_confidences_equal(self, other):
        return self._is_confidence_equal(self.understanding_confidence,
                                         other.understanding_confidence) and self._is_confidence_equal(
                                             self.weighted_understanding_confidence,
                                             other.weighted_understanding_confidence
                                         )

    def _are_perception_confidences_equal(self, other):
        return self._is_confidence_equal(self.perception_confidence, other.perception_confidence)

    def class_internal_move_content_equals(self, other):
        return True

    def move_content_equals(self, other):
        try:
            return (self.get_type() == other.get_type() and self.class_internal_move_content_equals(other))
        except AttributeError:
            return False

    def __ne__(self, other):
        return not (self == other)

    def is_icm(self):
        return False

    def is_question_raising(self):
        return self.get_type() == Move.ASK

    def is_turn_yielding(self):
        return self.get_type() == Move.ANSWER

    def to_rich_string(self):
        return "%s:%s:%s" % (unicode(self), self._speaker, self._understanding_confidence)

    def uprank(self, amount):
        self._weighted_understanding_confidence *= (1 + amount)

    def downrank(self, amount):
        self._weighted_understanding_confidence *= (1 - amount)

    def set_background(self, background):
        self._background = background

    def as_json(self):
        return {
            self.get_type(): {
                "ddd_name": convert_to_json(self._ddd_name),
                "speaker": convert_to_json(self._speaker),
                "understanding_confidence": convert_to_json(self._understanding_confidence),
                "weighted_understanding_confidence": convert_to_json(self._weighted_understanding_confidence),
                "perception_confidence": convert_to_json(self.perception_confidence),
                "modality": convert_to_json(self._modality),
                "utterance": convert_to_json(self._utterance),
                "background": convert_to_json(self._background),
            }
        }


class MoveWithSemanticContent(Move, SemanticObjectWithContent):
    def __init__(self, type, content, *args, **kwargs):
        Move.__init__(self, type, *args, **kwargs)
        SemanticObjectWithContent.__init__(self, content)
        self._content = content

    def get_content(self):
        return self._content

    def __hash__(self):
        return hash((
            self.__class__.__name__, self._type, self._perception_confidence, self._understanding_confidence,
            self._weighted_understanding_confidence, self._speaker, self._utterance, self._ddd_name
        ))

    def _get_expression(self, include_attributes):
        string = "Move("
        string += "%s(%s" % (self._type, unicode(self._content))
        if self._background:
            string += ", %s" % unicodify(self._background)
        string += ")"
        if include_attributes:
            string += self._build_string_from_attributes()
        string += ")"
        return string

    def class_internal_move_content_equals(self, other):
        return self._content == other._content

    def as_json(self):
        json = Move.as_json(self)
        json[self.get_type()]["content"] = convert_to_json(self._content)
        return json


class ICMMove(Move):
    RERAISE = "reraise"
    PER = "per"
    ACC = "acc"
    SEM = "sem"
    UND = "und"
    ACCOMMODATE = "accommodate"
    LOADPLAN = "loadplan"
    RESUME = "resume"

    INT = "int"
    POS = "pos"
    NEG = "neg"

    def __init__(
        self,
        type,
        understanding_confidence=None,
        speaker=None,
        polarity=None,
        ddd_name=None,
        perception_confidence=None
    ):
        if polarity is None:
            polarity = ICMMove.POS
        self._polarity = polarity
        Move.__init__(
            self,
            type,
            understanding_confidence=understanding_confidence,
            speaker=speaker,
            ddd_name=ddd_name,
            perception_confidence=perception_confidence
        )

    def class_internal_move_content_equals(self, other):
        return (self._polarity == other._polarity and self._type == other._type)

    def get_polarity(self):
        return self._polarity

    def is_icm(self):
        return True

    def is_issue_icm(self):
        return False

    def is_question_raising(self):
        return False

    def is_negative_perception_icm(self):
        if self.get_type() == ICMMove.PER:
            return self.get_polarity() == ICMMove.NEG
        else:
            return False

    def is_positive_acceptance_icm(self):
        if self.get_type() == ICMMove.ACC:
            return self.get_polarity() == ICMMove.POS
        else:
            return False

    def is_negative_acceptance_issue_icm(self):
        return False

    def is_negative_acceptance_icm(self):
        if (self.get_type() == ICMMove.ACC and self.get_polarity() == ICMMove.NEG):
            return True
        else:
            return False

    def as_json(self):
        self._icm_to_string()

    def get_semantic_expression(self):
        string = "ICMMove(%s" % self._icm_to_string()
        if self._speaker:
            string += ", speaker=%s" % self._speaker
        if self._understanding_confidence is not None:
            string += ", understanding_confidence=%s" % self._understanding_confidence
        if self.perception_confidence is not None:
            string += ", perception_confidence=%s" % self.perception_confidence
        string += ")"
        return string

    def is_negative_understanding_icm(self):
        return (self.get_type() == ICMMove.UND and self.get_polarity() == ICMMove.NEG)

    def is_positive_understanding_icm_with_non_neg_content(self):
        return False

    def is_interrogative_understanding_icm_with_non_neg_content(self):
        return False

    def is_turn_yielding(self):
        return self.get_type() == ICMMove.ACC and self._polarity == ICMMove.NEG

    def __unicode__(self):
        string = "ICMMove(%s" % self._icm_to_string()
        if self._speaker:
            string += ", speaker=%s" % self._speaker
        if self._understanding_confidence is not None:
            string += ", understanding_confidence=%s" % self._understanding_confidence
        if self.perception_confidence is not None:
            string += ", perception_confidence=%s" % self.perception_confidence
        string += ")"
        return string

    def _icm_to_string(self):
        if self._type in [ICMMove.PER, ICMMove.ACC, ICMMove.UND, ICMMove.SEM]:
            return "icm:%s*%s" % (self._type, self._polarity)
        return "icm:%s" % self._type


class IssueICMMove(ICMMove):
    def is_issue_icm(self):
        return True

    def is_negative_acceptance_issue_icm(self):
        if (self.get_type() == ICMMove.ACC and self.get_polarity() == ICMMove.NEG):
            return True
        return False

    def _icm_to_string(self):
        return "%s:issue" % ICMMove._icm_to_string(self)


class ICMMoveWithContent(ICMMove):
    def __init__(self, type, content, content_speaker=None, *args, **kwargs):
        ICMMove.__init__(self, type, *args, **kwargs)
        self._content = content
        self._content_speaker = self._get_checked_content_speaker(content_speaker)

    def get_content(self):
        return self._content

    def _get_checked_content_speaker(self, speaker):
        if (speaker in [Speaker.USR, Speaker.SYS, Speaker.MODEL, None]):
            return speaker
        raise Exception("'%s' is not a valid value for content_speaker" % speaker)

    def get_content_speaker(self):
        return self._content_speaker

    def _icm_to_string(self):
        if self._content_speaker is not None:
            return "icm:%s*%s:%s*%s" % (self._type, self._polarity, self._content_speaker, self._content)

        if self._type == ICMMove.PER:
            return 'icm:%s*%s:"%s"' % (self._type, self._polarity, self._content)
        if self._type in [ICMMove.ACC, ICMMove.UND, ICMMove.SEM]:
            return "icm:%s*%s:%s" % (self._type, self._polarity, self._content)
        return "icm:%s:%s" % (self._type, self._content)

    def class_internal_move_content_equals(self, other):
        return (
            self._polarity == other._polarity and self._type == other._type and self._content == other._content
            and self._content_speaker == other._content_speaker
        )

    def is_question_raising(self):
        return (
            self.get_type() == ICMMove.UND and self.get_content() is not None
            and not (self.get_polarity() == ICMMove.POS and not self.get_content().is_positive())
        )

    def is_positive_understanding_icm_with_non_neg_content(self):
        return (
            self.get_type() == ICMMove.UND and self.get_polarity() == ICMMove.POS and self.get_content().is_positive()
        )

    def is_interrogative_understanding_icm_with_non_neg_content(self):
        return (
            self.get_type() == ICMMove.UND and self.get_polarity() == ICMMove.INT and self.get_content().is_positive()
        )


class ICMMoveWithStringContent(ICMMoveWithContent):
    pass


class ICMMoveWithSemanticContent(ICMMoveWithContent, MoveWithSemanticContent):
    def __init__(
        self,
        type,
        content,
        understanding_confidence=None,
        speaker=None,
        ddd_name=None,
        content_speaker=None,
        polarity=None,
        perception_confidence=None
    ):
        MoveWithSemanticContent.__init__(
            self,
            type,
            content,
            understanding_confidence=understanding_confidence,
            speaker=speaker,
            ddd_name=ddd_name,
            perception_confidence=perception_confidence
        )
        ICMMoveWithContent.__init__(
            self,
            type,
            content,
            content_speaker=content_speaker,
            understanding_confidence=understanding_confidence,
            speaker=speaker,
            polarity=polarity,
            ddd_name=ddd_name,
            perception_confidence=perception_confidence
        )


class ReportMove(MoveWithSemanticContent):
    def __init__(self, report_proposition):
        self.report_proposition = report_proposition
        MoveWithSemanticContent.__init__(self, Move.REPORT, report_proposition)

    def is_turn_yielding(self):
        return True

    def _get_expression(self, include_attributes):
        string = "report(%s" % unicodify(self.report_proposition)
        if self._background:
            string += ", %s" % unicodify(self._background)
        if include_attributes:
            string += self._build_string_from_attributes()
        string += ")"
        return string

    def class_internal_move_content_equals(self, other):
        return (self.report_proposition == other.report_proposition)


class PrereportMove(Move, OntologySpecificSemanticObject):
    def __init__(self, ontology_name, service_action, arguments):
        self.service_action = service_action
        self.arguments = arguments
        Move.__init__(self, Move.PREREPORT)
        OntologySpecificSemanticObject.__init__(self, ontology_name)

    def get_service_action(self):
        return self.service_action

    def get_arguments(self):
        return self.arguments

    def __unicode__(self):
        return "prereport(%s, %s%s)" % (
            self.service_action, unicodify(self.arguments), self._build_string_from_attributes()
        )

    def class_internal_move_content_equals(self, other):
        return (self.service_action == other.service_action and self.arguments == other.arguments)


class NoMove(Move):
    def __init__(self, *args, **kwargs):
        super(NoMove, self).__init__(Move.NO_MOVE, *args, **kwargs)


class GreetMove(Move):
    def __init__(self, *args, **kwargs):
        super(GreetMove, self).__init__(Move.GREET, *args, **kwargs)


class MuteMove(Move):
    def __init__(self, *args, **kwargs):
        super(MuteMove, self).__init__(Move.MUTE, *args, **kwargs)


class UnmuteMove(Move):
    def __init__(self, *args, **kwargs):
        super(UnmuteMove, self).__init__(Move.UNMUTE, *args, **kwargs)


class QuitMove(Move):
    def __init__(self, *args, **kwargs):
        super(QuitMove, self).__init__(Move.QUIT, *args, **kwargs)


class AskMove(MoveWithSemanticContent):
    def __init__(self, question, *args, **kwargs):
        MoveWithSemanticContent.__init__(self, Move.ASK, question, *args, **kwargs)


class OntologySpecificAnswerMove(MoveWithSemanticContent):
    def __init__(self, answer, *args, **kwargs):
        MoveWithSemanticContent.__init__(self, Move.ANSWER, answer, *args, **kwargs)


class AnswerMove(MoveWithSemanticContent):
    def __init__(self, answer, *args, **kwargs):
        MoveWithSemanticContent.__init__(self, Move.ANSWER, answer, *args, **kwargs)


class RequestMove(MoveWithSemanticContent):
    def __init__(self, action, *args, **kwargs):
        MoveWithSemanticContent.__init__(self, Move.REQUEST, action, *args, **kwargs)



import unittest

import tala.model.set
from tala.model.tis_node import TISNode


class TISNodeTests(unittest.TestCase):
    def setUp(self):
        class TestNode(TISNode):
            def __init__(self):
                self.child1 = tala.model.set.Set()
                self.child2 = "string"

        class RecursiveTISNode(TISNode):
            def __init__(self):
                self.node1 = TestNode()
                self.node2 = TestNode()

        self.node = TestNode()
        self.recursive_node = RecursiveTISNode()

    def testTISNode_pretty_string_with_indent0(self):
        expected_string = "\n\tchild1: {}\n\tchild2: string"
        self.assertEquals(expected_string, self.node.pretty_string(0))

    def testTISNode_pretty_string_with_indent1(self):
        expected_string = "\n\t\tchild1: {}\n\t\tchild2: string"
        self.assertEquals(expected_string, self.node.pretty_string(1))

    def testTISNode_recursive_pretty_string(self):
        expected_string = "\n\tnode1: \n\t\tchild1: {}\n\t\tchild2: string\n\tnode2: \n\t\tchild1: {}\n\t\tchild2: string"
        self.assertEquals(expected_string, self.recursive_node.pretty_string(0))

    def testTISNode_pretty_string_base_case(self):
        expected_string = "\nchild1: {}\nchild2: string"
        self.assertEquals(expected_string, self.node.pretty_string_new([]))

    def testTISNode_pretty_string_recursive_case(self):
        expected_string = "\nnode1.child1: {}\nnode1.child2: string\nnode2.child1: {}\nnode2.child2: string"
        self.assertEquals(expected_string, self.recursive_node.pretty_string_new([]))



import os
import shutil
from StringIO import StringIO
import warnings

from tala.config import DddConfig
from tala.utils import chdir

DDD_MAKER_PATH = os.path.dirname(os.path.abspath(__file__))
TEMPLATES_PATH = os.path.join(DDD_MAKER_PATH, "templates")


def write_template_to_file(target_path, template):
    with open(target_path, 'w') as target:
        template.seek(0)
        shutil.copyfileobj(template, target)


def word_list_template():
    content = StringIO()
    with open(os.path.join(TEMPLATES_PATH, "word_list_template.txt"), 'r') as template:
        shutil.copyfileobj(template, content)
    return content


def word_list_filename(path_to_ddd):
    with chdir.chdir(path_to_ddd):
        ddd_config = DddConfig().read()
    filename = ddd_config["word_list"]
    return filename


def create_word_list_boilerplate(path_to_ddd):
    filename = word_list_filename(path_to_ddd)
    target_path = os.path.join(path_to_ddd, filename)
    template = word_list_template()
    write_template_to_file(target_path, template)


def potentially_create_word_list_boilerplate(path_to_ddd):
    filename = word_list_filename(path_to_ddd)
    word_list_path = os.path.join(path_to_ddd, filename)
    if not os.path.exists(word_list_path):
        create_word_list_boilerplate(path_to_ddd)
        ddd_name = os.path.basename(path_to_ddd)
        message = "The word list for %r is missing. It was created at %s." % (ddd_name, word_list_path)
        warnings.warn(message)



import warnings

from tala.model.grammar.intent import Question, Request, Answer
from tala.model.grammar.required_entity import RequiredPropositionalEntity, RequiredSortalEntity
from tala.model.sort import STRING
from tala.nl.gf import rgl_grammar_entry_types
from tala.nl.gf.grammar_entry_types import Constants
from tala.utils.as_json import JSONLoggable


class NoIndividualsFoundException(Exception):
    pass


class NoRequestsFoundException(Exception):
    pass


class NoQuestionsFoundException(Exception):
    pass


class UnexpectedIndividualsFoundException(Exception):
    pass


class UnexpectedRequestsFoundException(Exception):
    pass


class UnexpectedQuestionsFoundException(Exception):
    pass


class UnexpectedAnswersFoundException(Exception):
    pass


class UnexpectedAnswerFormatException(Exception):
    pass


class UnexpectedStringsFoundException(Exception):
    pass


class GrammarBase(JSONLoggable):
    def __init__(self, grammar_root, grammar_path):
        super(GrammarBase, self).__init__()
        self._grammar_root = grammar_root
        self._grammar_path = grammar_path
        self._local_individual_identifier = None

    def as_json(self):
        return {
            "answers": self.answers(),
        }

    def requests_of_action(self, action):
        raise NotImplementedError("%s.requests_of_action(...) need to be implemented." % self.__class__.__name__)

    def _find_individuals_of(self, name):
        raise NotImplementedError("%s._find_individual(...) need to be implemented." % self.__class__.__name__)

    def entries_of_individual(self, name):
        individuals = self._find_individuals_of(name)
        if len(individuals) < 1:
            all_names = [individual.get("name") for individual in self._grammar_root.findall(Constants.INDIVIDUAL)]
            raise NoIndividualsFoundException(
                "Expected at least one <%s ...> for individual '%s', but it was "
                "not found among %s in %s" % (Constants.INDIVIDUAL, name, all_names, self._grammar_path)
            )
        if len(individuals) > 1:
            raise UnexpectedIndividualsFoundException(
                "Expected a single <%s ...> for individual '%s' but found %d in %s" %
                (Constants.INDIVIDUAL, name, len(individuals), self._grammar_path)
            )
        return self._process_individual_entries(individuals[0])

    def _process_individual_entries(self, individual_element):
        if self._element_has_children(individual_element):
            return self._get_item_contents_for_individual_element(individual_element)
        else:
            return [individual_element.text]

    def _element_has_children(self, element):
        return len(list(element)) != 0

    def _get_item_contents_for_individual_element(self, element):
        one_of = element.find("one-of")
        return [e.text for e in list(one_of.iter(tag="item"))]

    def questions_of_predicate(self, predicate):
        questions = self._grammar_root.findall("question[@predicate='%s'][@speaker='user']" % (predicate))
        if len(questions) < 1:
            actual_questions = self._grammar_root.findall("question[@speaker='user']")
            actual_predicates = [question.get("predicate") for question in actual_questions]
            raise NoQuestionsFoundException(
                "Expected at least one <question speaker='user'...> for predicate '%s' but "
                "it was not found among %s in %s" % (predicate, actual_predicates, self._grammar_path)
            )
        if len(questions) > 1:
            raise UnexpectedQuestionsFoundException(
                "Expected a single <question speaker='user' ...> for predicate "
                "'%s' but found %d in %s" % (predicate, len(questions), self._grammar_path)
            )
        question = questions[0]
        items = self._plain_text_items_of(question)
        for item in items:
            text_chunks, required_entities = self._chunks_and_entities_of_item(item)
            yield Question(predicate, text_chunks, required_entities)

    def answers(self):
        answers = self._grammar_root.findall("answer[@speaker='user']")
        if not answers:
            return
        if len(answers) > 1:
            raise UnexpectedAnswersFoundException(
                "Expected a single <answer speaker='user'> but found %d in %s" % (len(answers), self._grammar_path)
            )
        answer = answers[0]
        items = self._plain_text_items_of(answer)
        for item in items:
            text_chunks, required_entities = self._chunks_and_entities_of_item(item)
            if not required_entities:
                raise UnexpectedAnswerFormatException(
                    "Expected at least one <%s .../> in every item in <answer speaker=\"user\"> "
                    "but found some without" % self._local_individual_identifier
                )
            yield Answer(text_chunks, required_entities)

    def strings_of_predicate(self, predicate):
        strings = self._grammar_root.findall("string[@predicate='%s']" % predicate)
        if len(strings) > 1:
            raise UnexpectedStringsFoundException(
                "Expected a single <string predicate='%s'> but found %d in %s" %
                (predicate, len(strings), self._grammar_path)
            )

        if len(strings) == 0:
            warnings.warn(
                "Expected training examples for predicate '{predicate}' of sort '{sort}' but found none. "
                "Add them with:\n"
                "\n"
                "<string predicate=\"{predicate}\">\n"
                "  <one-of>\n"
                "    <item>an example</item>\n"
                "    <item>another example</item>\n"
                "  </one-of>\n"
                "</string>".format(predicate=predicate, sort=STRING)
            )
            return

        string = strings[0]
        items = self._plain_text_items_of(string)
        for item in items:
            text_chunks, required_entities = self._chunks_and_entities_of_item(item)
            if any(required_entities):
                raise UnexpectedIndividualsFoundException(
                    "Expected no <%s ...> in <string predicate='%s'> but found some" %
                    (self._local_individual_identifier, predicate)
                )
            assert len(text_chunks) == 1, "Expected 1 text chunk but got %d" % len(text_chunks)
            yield text_chunks[0]

    def _chunks_and_entities_of_item(self, item):
        text_chunks = [item.text or ""]
        required_entities = []
        local_individuals = item.findall(self._local_individual_identifier)
        for individual_element in local_individuals:
            required_entity = self._required_entity_from_individual_element(individual_element)
            required_entities.append(required_entity)
            tail = individual_element.tail or ""
            text_chunks.append(tail)
        self._strip_chunks(text_chunks)
        return text_chunks, required_entities

    def _strip_chunks(self, chunks):
        chunks[0] = chunks[0].lstrip()
        chunks[-1] = chunks[-1].rstrip()

    def _required_entity_from_individual_element(self, element):
        predicate = element.attrib.get("predicate")
        if predicate:
            return RequiredPropositionalEntity(predicate)
        sort = element.attrib.get("sort")
        if sort:
            return RequiredSortalEntity(sort)
        raise UnexpectedIndividualsFoundException(
            "Expected either a 'sort' or 'predicate' attribute in "
            "<%s ...>: %s" % (self._local_individual_identifier, element)
        )

    def _plain_text_items_of(self, item):
        raise NotImplementedError("%s._plain_text_items_of(...) need to be implemented." % self.__class__.__name__)


class Grammar(GrammarBase):
    def __init__(self, language_code, grammar_path):
        super(Grammar, self).__init__(language_code, grammar_path)
        self._local_individual_identifier = Constants.SLOT

    def requests_of_action(self, name):
        actions = self._grammar_root.findall("%s[@name='%s']" % (Constants.ACTION, name))
        if len(actions) < 1:
            all_actions = self._grammar_root.findall(Constants.ACTION)
            all_names = [action.get("name") for action in all_actions]
            raise NoRequestsFoundException(
                "Expected at least one <action ...> for action '%s' but it was not found "
                "among %s in %s" % (name, all_names, self._grammar_path)
            )
        if len(actions) > 1:
            raise UnexpectedRequestsFoundException(
                "Expected a single <action ...> for action '%s' but found %d in %s" %
                (name, len(actions), self._grammar_path)
            )
        action_item = actions[0]
        items = list(self._plain_text_items_of(action_item))
        for item in items:
            text_chunks, required_entities = self._chunks_and_entities_of_item(item)
            yield Request(name, text_chunks, required_entities)

    def _plain_text_items_of(self, root):
        def get_items(root):
            items = self._items_of(root)
            for item in items:
                if item.find(Constants.VP) is not None or item.find(Constants.NP) is not None:
                    continue
                yield item

        plain_text_items = list(get_items(root))
        if not plain_text_items:
            warnings.warn(
                "%s ignores element '%s' with attributes %s since there are no plain text items" %
                (self.__class__.__name__, root.tag, root.attrib)
            )
        return plain_text_items

    def _items_of(self, root):
        items = root.findall("./%s/%s" % (Constants.ONE_OF, Constants.ITEM))
        if not items:
            yield root
        for item in items:
            yield item

    def _find_individuals_of(self, name):
        return self._grammar_root.findall("%s[@name='%s']" % (Constants.INDIVIDUAL, name))


class GrammarForRGL(GrammarBase):
    def __init__(self, grammar_root, grammar_path):
        super(GrammarForRGL, self).__init__(grammar_root, grammar_path)
        self._local_individual_identifier = Constants.INDIVIDUAL

    def requests_of_action(self, action):
        requests = self._grammar_root.findall("%s[@action='%s']" % (rgl_grammar_entry_types.REQUEST, action))
        if len(requests) < 1:
            actual_requests = self._grammar_root.findall(rgl_grammar_entry_types.REQUEST)
            actual_actions = [request.get("action") for request in actual_requests]
            raise NoRequestsFoundException(
                "Expected at least one <request ...> for action '%s' but it was not found "
                "among %s in %s" % (action, actual_actions, self._grammar_path)
            )
        if len(requests) > 1:
            raise UnexpectedRequestsFoundException(
                "Expected a single <request ...> for action '%s' but found %d in %s" %
                (action, len(requests), self._grammar_path)
            )
        request = requests[0]
        items = self._plain_text_items_of(request)
        for item in items:
            text_chunks, required_entities = self._chunks_and_entities_of_item(item)
            yield Request(action, text_chunks, required_entities)

    def _plain_text_items_of(self, item):
        utterances = item.findall(rgl_grammar_entry_types.UTTERANCE)
        if not utterances:
            warnings.warn(
                "%s ignores element '%s' with attributes %s since it has no <utterance>" %
                (self.__class__.__name__, item.tag, item.attrib)
            )
            return

        for utterance in utterances:
            items = utterance.findall("./%s/%s" % (Constants.ONE_OF, Constants.ITEM))
            if not items:
                yield utterance
            for item in items:
                yield item

    def _find_individuals_of(self, name):
        return self._grammar_root.findall(
            "%s[@name='%s']/%s" % (Constants.INDIVIDUAL, name, rgl_grammar_entry_types.PROPER_NOUN)
        )



import os
import os.path
import shutil
import tempfile

from mock import Mock

from tala.ddd.building.steps.generate import AbstractGenerateStep
from tala.model.ddd import DDD


class TestAbstractGenerateStep(object):
    def setup(self):
        self._temp_dir = tempfile.mkdtemp(prefix="TestGenerate")
        self._cwd = os.getcwd()
        os.chdir(self._temp_dir)
        self._generate_step = None
        self._mock_ddd = None
        self._ddd_name = None
        self._language_codes = []

    def teardown(self):
        os.chdir(self._cwd)
        shutil.rmtree(self._temp_dir)

    def test_build_generates_grammars(self):
        self._given_ddd_name("MockDDD")
        self._given_mocked_ddd(is_rasa_enabled=False)
        self._given_language_codes(["eng"])
        self._given_file_exists("MockDDD/grammar/grammar_eng.py")
        self._given_generate_step_created()
        self._given_mocked_generate_grammars_method()
        self._when_calling_build()
        self._then_generate_grammars_is_invoked_with_arguments("eng")

    def _given_ddd_name(self, name):
        self._ddd_name = name

    def _given_mocked_ddd(self, is_rasa_enabled):
        mock_ddd = Mock(spec=DDD)
        mock_ddd.name = self._ddd_name
        mock_ddd.is_rasa_enabled = is_rasa_enabled
        self._mock_ddd = mock_ddd

    def _given_language_codes(self, language_codes):
        self._language_codes = language_codes

    def _given_file_exists(self, path):
        self._ensure_dir_exists(os.path.dirname(path))
        with open(path, "w"):
            pass

    def _ensure_dir_exists(self, path):
        if not os.path.exists(path):
            os.makedirs(path)

    def _given_generate_step_created(self):
        ignore_warnings = True
        verbose = False
        self._generate_step = AbstractGenerateStep(
            self._mock_ddd, ignore_warnings, self._language_codes, verbose, self._ddd_name, "grammar"
        )

    def _given_mocked_generate_grammars_method(self):
        self._generate_step._generate_grammars = Mock()

    def _when_calling_build(self):
        self._generate_step.build()

    def _then_generate_grammars_is_invoked_with_arguments(self, *expected_args):
        self._generate_step._generate_grammars.assert_called_once_with(*expected_args)



import copy

from tala.model.error import OntologyError
from tala.model.goal import PerformGoal
from tala.model.lambda_abstraction import LambdaAbstractedPredicateProposition
from tala.model.proposition import PropositionSet, GoalProposition
from tala.model.question import AltQuestion, WhQuestion
from tala.model.test.utils import LibTestCase


class WhQuestionTests(LibTestCase):
    def setUp(self):
        self.setUpLibTestCase()
        self.lambda_abstracted_dest_city = LambdaAbstractedPredicateProposition(
            self.predicate_dest_city, self.ontology.get_name()
        )
        self.dest_city_wh_question = WhQuestion(self.lambda_abstracted_dest_city)

    def test_is_question(self):
        self.assertTrue(self.dest_city_wh_question.is_question())

    def test_string_format_predicate(self):
        self.assertEquals("?X.dest_city(X)", unicode(self.dest_city_wh_question))

    def test_string_format_goal(self):
        goal_question = WhQuestion(self.lambda_abstracted_goal_prop)
        self.assertEquals("?X.goal(X)", unicode(goal_question))

    def test_wh_question_get_content(self):
        self.assertEquals(self.lambda_abstracted_dest_city, self.dest_city_wh_question.get_content())

    def testWhQuestionEquality(self):
        question = self.dest_city_question
        identical_question = copy.copy(self.dest_city_question)
        self.assert_eq_returns_true_and_ne_returns_false_symmetrically(question, identical_question)

    def testWhQuestionNonEquality(self):
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(self.dest_city_question, self.price_question)

    def test_is_wh_question(self):
        self.assertTrue(self.dest_city_wh_question.is_wh_question())


class QuestionTests(LibTestCase):
    def setUp(self):
        self.setUpLibTestCase()

    def test_is_question(self):
        question = self.price_question
        non_question = self.proposition_dest_city_paris
        self.assertTrue(question.is_question())
        self.assertFalse(non_question.is_question())

    def test_questions_are_hashable(self):
        set([self.price_question])


class YesNoQuestionTests(LibTestCase):
    def setUp(self):
        self.setUpLibTestCase()
        self.question = self.ontology.create_yes_no_question("dest_city", "paris")

    def test_is_question(self):
        self.assertTrue(self.question.is_question())

    def testYesNoQuestionunicode(self):
        question = self.ontology.create_yes_no_question("dest_city", "paris")
        self.assertEquals("?dest_city(paris)", unicode(question))

    def testInvalidYesNoQuestion(self):
        with self.assertRaises(OntologyError):
            self.ontology.create_yes_no_question("kalle", "paris")


class AltQuestionTests(LibTestCase):
    def setUp(self):
        self.setUpLibTestCase()
        propositions = [self.proposition_dest_city_paris, self.proposition_dest_city_london]
        self.question = AltQuestion(PropositionSet(propositions))

    def test_is_question(self):
        self.assertTrue(self.question.is_question())

    def test_is_alt_question(self):
        self.assertTrue(self.question.is_alt_question())

    def test_alt_question_equality(self):
        identical_question = copy.copy(self.question)
        self.assert_eq_returns_true_and_ne_returns_false_symmetrically(self.question, identical_question)

    def test_alt_question_inequality(self):
        other_propositions = PropositionSet([self.proposition_dest_city_paris])
        non_identical_question = AltQuestion(other_propositions)
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(self.question, non_identical_question)

    def test_string_format_for_goal_alt_question(self):
        alt_question = AltQuestion(
            PropositionSet([
                GoalProposition(PerformGoal(self.top_action)),
                GoalProposition(PerformGoal(self.buy_action))
            ])
        )
        self.assertEquals("?set([goal(perform(top)), goal(perform(buy))])", unicode(alt_question))

    def test_string_format_for_predicate_alt_question(self):
        self.assertEquals("?X.dest_city(X), set([dest_city(paris), dest_city(london)])", unicode(self.question))



# -*- coding: utf-8 -*-

from mock import Mock, patch

import tala.ddd.ddd_xml_compiler
from tala.ddd.ddd_xml_compiler import DddXmlCompiler, DddXmlCompilerException, ViolatesSchemaException, UnexpectedAttributeException
from tala.ddd.parser import Parser
from tala.ddd.services.service_interface import ServiceParameter, ServiceActionInterface, ServiceEntityRecognizerInterface, ServiceQueryInterface, ServiceValidatorInterface, FrontendTarget, DeviceModuleTarget, HttpTarget, ServiceInterface, ActionFailureReason, PlayAudioActionInterface, AudioURLServiceParameter
from tala.ddd.test.ddd_compiler_test_case import DddCompilerTestCase
from tala.model.device import ParameterField
from tala.model.domain import Domain
from tala.model.ontology import Ontology
from tala.model.plan import Plan
from tala.model.plan_item import IfThenElse, ForgetAllPlanItem, InvokeServiceQueryPlanItem, InvokeServiceActionPlanItem
from tala.model.predicate import Predicate
from tala.model.sort import CustomSort, RealSort, UndefinedSort
from tala.nl.gf import rgl_grammar_entry_types as rgl_types
from tala.nl.gf.grammar_entry_types import Node, Constants


class DddXmlCompilerTestCase(DddCompilerTestCase):
    def _given_compiled_ontology(
        self, xml='<ontology name="MockupOntology"/>', domain_name="Domain", ddd_name="mock_ddd"
    ):
        self._ddd_name = ddd_name
        ontology_args = DddXmlCompiler().compile_ontology(xml)
        self._ontology = Ontology(**ontology_args)
        self._parser = Parser(self._ddd_name, self._ontology, domain_name)
        self._service_interface = Mock(spec=ServiceInterface)

    def _when_compile_ontology(self, ontology_xml):
        self._compile_ontology(ontology_xml)

    def _compile_ontology(self, ontology_xml):
        self._result = DddXmlCompiler().compile_ontology(ontology_xml)
        self._ontology = Ontology(**self._result)

    def _predicate(self, *args, **kwargs):
        return Predicate(self._ontology.name, *args, **kwargs)

    def _when_compile_ontology_then_exception_is_raised(self, ontolog_xml, expected_exception, expected_message):
        with self.assertRaises(expected_exception) as context_manager:
            self._compile_ontology(ontolog_xml)
        self.assertEquals(expected_message, str(context_manager.exception))

    def _when_compile_domain(self, domain_xml='<domain name="Domain" />'):
        self._result = DddXmlCompiler().compile_domain(
            self._ddd_name, domain_xml, self._ontology, self._parser, self._service_interface
        )

    def _when_compile_domain_with_plan(self, plan_xml):
        self._when_compile_domain(
            """
<domain name="Domain">
  <goal type="perform" action="top">
    <plan>%s</plan>
  </goal>
</domain>""" % plan_xml
        )

    def _when_compile_plan_with_attribute(self, key, value):
        self._when_compile_domain(
            """
<domain name="Domain">
  <goal type="perform" action="top" %s="%s" />
</domain>""" % (key, value)
        )

    def _when_compile_plan_with_content(self, xml):
        self._when_compile_domain(
            """
<domain name="Domain">
  <goal type="perform" action="top">%s</goal>
</domain>""" % xml
        )

    def _then_result_has_keys(self, expected_keys):
        self.assertItemsEqual(expected_keys, self._result.keys())

    def _then_result_has_field(self, expected_key, expected_value):
        self.assertEquals(expected_value, self._result[expected_key])

    def _then_result_has_plan(self, expected_plan):
        actual_plans = self._result["plans"]
        actual_plan = actual_plans[0]["plan"]
        self.assertEquals(expected_plan, actual_plan)

    def _then_result_has_plan_with_attribute(self, expected_key, expected_value):
        actual_plans = self._result["plans"]
        actual_plan = actual_plans[0]
        self.assertEquals(expected_value, actual_plan[expected_key])

    def _parse(self, string):
        return self._parser.parse(string)

    def _when_compile_grammar(self, *args):
        self._compile_grammar(*args)

    def _compile_grammar(self, string):
        self._result = DddXmlCompiler().compile_grammar(string, self._ontology, self._device_handler, "eng")

    def _when_compile_rgl_grammar(self, string):
        self._result = DddXmlCompiler().compile_rgl_grammar(string, self._ontology, self._service_interface, "eng")

    def _then_result_is(self, expected_result):
        self.assertEquals(expected_result, self._result)

    def _then_grammar_is(self, expected_grammar_children):
        expected_grammar_node = Node(Constants.GRAMMAR, {}, expected_grammar_children)
        self.assertEquals(
            expected_grammar_node, self._result,
            "\n%s\n!=\n%s\n\n(%r)" % (expected_grammar_node, self._result, self._result)
        )

    def _mock_warnings(self):
        return patch("%s.warnings" % tala.ddd.ddd_xml_compiler.__name__)

    def _then_warning_is_issued(self, mocked_warnings, expected_message):
        mocked_warnings.warn.assert_called_once_with(expected_message)


class OntologyCompilerTests(DddXmlCompilerTestCase):
    def test_name(self):
        self._when_compile_ontology('<ontology name="MockupOntology"/>')
        self._then_result_has_field("name", "MockupOntology")

    def test_custom_sort(self):
        self._when_compile_ontology("""
<ontology name="Ontology">
  <sort name="city"/>
</ontology>""")
        self._then_result_has_field("sorts", set([CustomSort(self._ontology.name, "city")]))

    def test_dynamic_sort(self):
        self._when_compile_ontology("""
<ontology name="Ontology">
  <sort name="city" dynamic="true"/>
</ontology>""")
        self._then_result_has_field("sorts", set([CustomSort(self._ontology.name, "city", dynamic=True)]))

    def test_predicate_of_custom_sort(self):
        self._when_compile_ontology(
            """
<ontology name="Ontology">
  <predicate name="dest_city" sort="city"/>
  <sort name="city"/>
</ontology>"""
        )
        self._then_result_has_field(
            "predicates", set([self._predicate("dest_city", CustomSort(self._ontology.name, "city"))])
        )

    def test_predicate_of_builtin_sort(self):
        self._when_compile_ontology(
            """
<ontology name="Ontology">
  <predicate name="price" sort="real"/>
</ontology>"""
        )
        self._then_result_has_field("predicates", set([self._predicate("price", RealSort())]))

    def test_predicate_of_dynamic_sort(self):
        self._when_compile_ontology(
            """
<ontology name="Ontology">
  <predicate name="dest_city" sort="city"/>
  <sort name="city" dynamic="true"/>
</ontology>"""
        )
        self._then_result_has_field(
            "predicates", set([self._predicate("dest_city", CustomSort(self._ontology.name, "city", dynamic=True))])
        )

    def test_feature_of(self):
        self._when_compile_ontology(
            """
<ontology name="Ontology">
  <sort name="city"/>
  <predicate name="dest_city" sort="city"/>
  <predicate name="dest_city_type" sort="city" feature_of="dest_city"/>
</ontology>"""
        )
        self._then_result_has_field(
            "predicates",
            set([
                self._predicate("dest_city", sort=CustomSort(self._ontology.name, "city")),
                self._predicate(
                    "dest_city_type", sort=CustomSort(self._ontology.name, "city"), feature_of_name="dest_city"
                )
            ])
        )

    def test_multiple_instances(self):
        self._when_compile_ontology(
            """
<ontology name="Ontology">
  <sort name="city"/>
  <predicate name="dest_city" sort="city" multiple_instances="true"/>
</ontology>"""
        )
        self._then_result_has_field(
            "predicates",
            set([self._predicate("dest_city", sort=CustomSort(self._ontology.name, "city"), multiple_instances=True)])
        )

    def test_individuals(self):
        self._when_compile_ontology(
            """
<ontology name="Ontology">
  <individual name="paris" sort="city"/>
  <sort name="city"/>
</ontology>"""
        )
        self._then_result_has_field("individuals", {"paris": CustomSort(self._ontology.name, "city")})

    def test_actions(self):
        self._when_compile_ontology("""
<ontology name="Ontology">
  <action name="buy"/>
</ontology>""")
        self._then_result_has_field("actions", set(["buy"]))

    def test_exception_for_undefined_sort_in_predicate(self):
        self._when_compile_ontology_then_exception_is_raised(
            """
<ontology name="Ontology">
  <predicate name="dest_city" sort="undefined_sort"/>
</ontology>""", UndefinedSort, "Expected a defined sort but got 'undefined_sort'."
        )


class PlanCompilationTests(DddXmlCompilerTestCase):
    def test_plan_for_perform_goal(self):
        self._given_compiled_ontology()

        self._when_compile_domain("""
<domain name="Domain">
  <goal type="perform" action="top" />
</domain>""")

        self._then_result_has_field("plans", [{"goal": self._parse("perform(top)"), "plan": Plan([])}])

    def test_plan_for_resolve_goal_for_wh_question(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <predicate name="price" sort="real"/>
</ontology>"""
        )

        self._when_compile_domain(
            """
<domain name="Domain">
  <goal type="resolve" question_type="wh_question" predicate="price" />
</domain>"""
        )

        self._then_result_has_field("plans", [{"goal": self._parse("resolve(?X.price(X))"), "plan": Plan([])}])

    def test_plan_for_resolve_goal_for_nullary_predicate_proposition(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <predicate name="need_visa" sort="boolean"/>
</ontology>"""
        )

        self._when_compile_domain(
            """
<domain name="Domain">
  <goal type="resolve" question_type="yn_question">
    <proposition predicate="need_visa"/>
  </goal>
</domain>"""
        )

        self._then_result_has_field("plans", [{"goal": self._parse("resolve(?need_visa())"), "plan": Plan([])}])

    def test_plan_for_handle_goal(self):
        self._given_compiled_ontology()

        self._when_compile_domain(
            """
<domain name="Domain">
  <goal type="handle" action="MockupAction" />
</domain>"""
        )

        self._then_result_has_field("plans", [{"goal": self._parse("handle(MockupAction)"), "plan": Plan([])}])

    def test_exception_raised_for_goal_without_type(self):
        self._given_compiled_ontology()

        with self.assertRaises(DddXmlCompilerException):
            self._when_compile_domain("""
<domain name="Domain">
  <goal />
</domain>""")

    def test_plan_stack_order(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <sort name="how"/>
  <sort name="city"/>
  <predicate name="means_of_transport" sort="how"/>
  <predicate name="dest_city" sort="city"/>
</ontology>"""
        )

        self._when_compile_domain_with_plan(
            """
<findout type="wh_question" predicate="means_of_transport" />
<findout type="wh_question" predicate="dest_city" />"""
        )

        self._then_result_has_plan(
            Plan([self._parse("findout(?X.dest_city(X))"),
                  self._parse("findout(?X.means_of_transport(X))")])
        )

    def test_preferred_proposition(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <predicate name="dest_city" sort="city"/>
  <individual name="paris" sort="city"/>
  <sort name="city"/>
</ontology>"""
        )
        self._when_compile_plan_with_content(
            """
<preferred>
  <proposition predicate="dest_city" value="paris"/>
</preferred>
"""
        )
        self._then_result_has_plan_with_attribute("preferred", self._parse("dest_city(paris)"))

    def test_preferred_boolean(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <action name="apply_for_membership"/>
</ontology>"""
        )
        self._when_compile_plan_with_content("""
<preferred/>
""")
        self._then_result_has_plan_with_attribute("preferred", True)

    def test_accommodate_without_feedback(self):
        self._given_compiled_ontology()
        self._when_compile_plan_with_attribute("accommodate_without_feedback", "true")
        self._then_result_has_plan_with_attribute("accommodate_without_feedback", True)

    def test_dynamic_title(self):
        self._given_compiled_ontology()
        self._when_compile_plan_with_attribute("dynamic_title", "true")
        self._then_result_has_plan_with_attribute("dynamic_title", True)

    def test_restart_on_completion(self):
        self._given_compiled_ontology()
        self._when_compile_plan_with_attribute("restart_on_completion", "true")
        self._then_result_has_plan_with_attribute("restart_on_completion", True)

    def test_reraise_on_resume(self):
        self._given_compiled_ontology()
        self._when_compile_plan_with_attribute("reraise_on_resume", "true")
        self._then_result_has_plan_with_attribute("reraise_on_resume", True)

    def test_io_status(self):
        self._given_compiled_ontology()
        self._when_compile_plan_with_attribute("io_status", "disabled")
        self._then_result_has_plan_with_attribute("io_status", Domain.DISABLED_IO_STATUS)

    def test_gui_context(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <predicate name="price" sort="real"/>
</ontology>"""
        )
        self._when_compile_plan_with_content('<gui_context predicate="price" />')
        self._then_result_has_plan_with_attribute("gui_context", [self._ontology.get_predicate("price")])

    def test_postplan(self):
        self._given_compiled_ontology()
        self._when_compile_plan_with_content('<postplan><forget_all/></postplan>')
        self._then_result_has_plan_with_attribute("postplan", Plan([ForgetAllPlanItem()]))

    def test_postcond_with_whitespace(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <predicate name="dest_city" sort="city"/>
  <individual name="paris" sort="city"/>
  <sort name="city"/>
</ontology>"""
        )
        self._when_compile_plan_with_content(
            """
<postcond>
  <proposition predicate="dest_city" value="paris"/>
</postcond>"""
        )
        self._then_result_has_plan_with_attribute("postconds", [self._parse("dest_city(paris)")])

    def test_postcond_for_predicate_proposition(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <predicate name="dest_city" sort="city"/>
  <individual name="paris" sort="city"/>
  <sort name="city"/>
</ontology>"""
        )
        self._when_compile_plan_with_content('<postcond><proposition predicate="dest_city" value="paris"/></postcond>')
        self._then_result_has_plan_with_attribute("postconds", [self._parse("dest_city(paris)")])


class DomainCompilerTests(DddXmlCompilerTestCase):
    def test_name(self):
        self._given_compiled_ontology()
        self._when_compile_domain()
        self._then_result_has_field("name", "Domain")

    def test_dependency(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <sort name="city"/>
  <sort name="country"/>
  <predicate name="dest_city" sort="city"/>
  <predicate name="dest_country" sort="country"/>
</ontology>"""
        )

        self._when_compile_domain(
            """
<domain name="Domain">
  <dependency type="wh_question" predicate="dest_country">
    <question type="wh_question" predicate="dest_city" />
  </dependency>
</domain>"""
        )

        self._then_result_has_field(
            "dependencies", {self._parse("?X.dest_country(X)"): set([self._parse("?X.dest_city(X)")])}
        )

    def test_default_questions(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <predicate name="price" sort="real"/>
</ontology>"""
        )

        self._when_compile_domain(
            """
<domain name="Domain">
  <default_question type="wh_question" predicate="price" />
</domain>"""
        )

        self._then_result_has_field("default_questions", [self._parse("?X.price(X)")])

    def test_superactions(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <action name="top"/>
  <action name="buy"/>
</ontology>"""
        )

        self._when_compile_domain(
            """
<domain name="Domain">
  <goal type="perform" action="top" />
  <goal type="perform" action="buy">
    <superaction name="top" />
  </goal>
</domain>"""
        )

        self._then_result_has_field(
            "plans", [{
                "goal": self._parse("perform(top)"),
                "plan": Plan([])
            }, {
                "goal": self._parse("perform(buy)"),
                "plan": Plan([]),
                "superactions": [self._parse("top")]
            }]
        )

    def test_parameters(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <predicate name="price" sort="real"/>
</ontology>"""
        )

        self._when_compile_domain(
            """
<domain name="Domain">
  <parameters question_type="wh_question" predicate="price" graphical_type="list" />
</domain>"""
        )

        self._then_result_has_field("parameters", {self._parse("?X.price(X)"): {"graphical_type": "list"}})


class ParameterCompilationTestCase(DddXmlCompilerTestCase):
    def test_choice_parameter(self):
        self._given_compiled_ontology()
        self._when_compile_parameter_for_question("graphical_type", "list")
        self._then_result_has_parameter_for_question("graphical_type", "list")

    def test_boolean_parameter(self):
        self._given_compiled_ontology()
        self._when_compile_parameter_for_question("incremental", "true")
        self._then_result_has_parameter_for_question("incremental", True)

    def test_question_valued_parameter(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <sort name="how"/>
  <predicate name="means_of_transport" sort="how"/>
  <predicate name="available_means_of_transport" sort="how"/>
</ontology>"""
        )

        self._when_compile_domain(
            """
<domain name="Domain">
  <parameters question_type="wh_question" predicate="means_of_transport">
    <service_query type="wh_question" predicate="available_means_of_transport"/>
  </parameters>
</domain>"""
        )

        self._then_result_has_field(
            "parameters", {
                self._parse("?X.means_of_transport(X)"):
                self._parser.parse_parameters("{service_query=?X.available_means_of_transport(X)}")
            }
        )

    def test_questions_valued_parameter(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <sort name="product"/>
  <predicate name="product_to_add" sort="product"/>
  <predicate name="product_tag" sort="string"/>
  <predicate name="product_title" sort="string"/>
</ontology>"""
        )

        self._when_compile_domain(
            """
<domain name="Domain">
  <parameters question_type="wh_question" predicate="product_to_add">
    <label_question type="wh_question" predicate="product_tag"/>
    <label_question type="wh_question" predicate="product_title"/>
  </parameters>
</domain>"""
        )

        self._then_result_has_field(
            "parameters", {
                self._parse("?X.product_to_add(X)"):
                self._parser.parse_parameters("{label_questions=[?X.product_tag(X), ?X.product_title(X)]}")
            }
        )

    def test_background_for_wh_question(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <sort name="how"/>
  <sort name="city"/>
  <predicate name="means_of_transport" sort="how"/>
  <predicate name="dest_city" sort="city"/>
</ontology>"""
        )

        self._when_compile_domain(
            """
<domain name="Domain">
  <parameters question_type="wh_question" predicate="means_of_transport">
    <background predicate="dest_city"/>
  </parameters>
</domain>"""
        )

        self._then_result_has_field(
            "parameters",
            {self._parse("?X.means_of_transport(X)"): self._parser.parse_parameters("{background=[dest_city]}")}
        )

    def test_background_for_yn_question(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <sort name="city"/>
  <predicate name="price" sort="real"/>
  <predicate name="dest_city" sort="city"/>
</ontology>"""
        )

        self._when_compile_domain(
            """
<domain name="Domain">
  <parameters question_type="yn_question">
    <resolve type="wh_question" predicate="price"/>
    <background predicate="dest_city"/>
  </parameters>
</domain>"""
        )

        self._then_result_has_field(
            "parameters",
            {self._parse("?goal(resolve(?X.price(X)))"): self._parser.parse_parameters("{background=[dest_city]}")}
        )

    def test_ask_feature(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <sort name="city"/>
  <predicate name="dest_city" sort="city"/>
  <predicate name="dest_city_type" sort="city" feature_of="dest_city"/>
</ontology>"""
        )

        self._when_compile_domain(
            """
<domain name="Domain">
  <parameters question_type="wh_question" predicate="dest_city">
    <ask_feature predicate="dest_city_type"/>
  </parameters>
</domain>"""
        )

        self._then_result_has_field(
            "parameters",
            {self._parse("?X.dest_city(X)"): self._parser.parse_parameters("{ask_features=[dest_city_type]}")}
        )

    def test_predicate_alts_parameter(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <sort name="how"/>
  <predicate name="means_of_transport" sort="how"/>
  <individual name="bus" sort="how"/>
  <individual name="train" sort="how"/>
</ontology>"""
        )

        self._when_compile_domain(
            """
<domain name="Domain">
  <parameters question_type="wh_question" predicate="means_of_transport">
    <alt><proposition predicate="means_of_transport" value="bus"/></alt>
    <alt><proposition predicate="means_of_transport" value="train"/></alt>
  </parameters>
</domain>"""
        )

        self._then_result_has_field(
            "parameters", {
                self._parse("?X.means_of_transport(X)"):
                self._parser.parse_parameters("{alts=set([means_of_transport(bus), means_of_transport(train)])}")
            }
        )

    def test_empty_alt_yields_exception(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <sort name="how"/>
  <predicate name="means_of_transport" sort="how"/>
</ontology>"""
        )

        with self.assertRaises(DddXmlCompilerException):
            self._when_compile_domain(
                """
<domain name="Domain">
  <parameters question_type="wh_question" predicate="means_of_transport">
    <alt/>
  </parameters>
</domain>"""
            )

    def test_goal_alts_parameter(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <predicate name="price" sort="real"/>
  <action name="buy"/>
</ontology>"""
        )

        self._when_compile_domain(
            """
<domain name="Domain">
  <parameters question_type="goal">
    <alt><perform action="buy"/></alt>
    <alt><resolve type="wh_question" predicate="price"/></alt>
  </parameters>
</domain>"""
        )

        self._then_result_has_field(
            "parameters", {
                self._parse("?X.goal(X)"):
                self._parser.parse_parameters("{alts=set([goal(perform(buy)), goal(resolve(?X.price(X)))])}")
            }
        )

    def test_parameters_for_predicate(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <predicate name="price" sort="real"/>
  <predicate name="dest_city" sort="city"/>
  <sort name="city"/>
</ontology>"""
        )

        self._when_compile_domain(
            """
<domain name="Domain">
  <parameters question_type="wh_question" predicate="price">
    <related_information type="wh_question" predicate="dest_city"/>
  </parameters>
</domain>"""
        )

        self._then_result_has_field(
            "parameters",
            {self._parse("?X.price(X)"): self._parser.parse_parameters("{related_information=[?X.dest_city(X)]}")}
        )

    def _given_compiled_ontology(self, xml=None):
        if xml is None:
            xml = """
<ontology name="Ontology">
  <predicate name="price" sort="real"/>
</ontology>"""
        DddXmlCompilerTestCase._given_compiled_ontology(self, xml)

    def _when_compile_parameter_for_question(self, name, value):
        self._when_compile_domain(
            """
<domain name="Domain">
  <parameters question_type="wh_question" predicate="price" %s="%s" />
</domain>""" % (name, value)
        )

    def _then_result_has_parameter_for_question(self, expected_name, expected_value):
        self._then_result_has_field("parameters", {self._parse("?X.price(X)"): {expected_name: expected_value}})


class PlanItemCompilationTestCase(DddXmlCompilerTestCase):
    def test_findout_for_wh_question(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <sort name="how"/>
  <predicate name="means_of_transport" sort="how"/>
</ontology>"""
        )

        self._when_compile_domain_with_plan('<findout type="wh_question" predicate="means_of_transport" />')

        self._then_result_has_plan(Plan([self._parse("findout(?X.means_of_transport(X))")]))

    def test_findout_for_goal_alt_question(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <action name="action1"/>
  <action name="action2"/>
</ontology>"""
        )

        self._when_compile_domain_with_plan(
            """
<findout type="alt_question">
  <alt><perform action="action1"/></alt>
  <alt><perform action="action2"/></alt>
</findout>"""
        )

        self._then_result_has_plan(
            Plan([self._parse("findout(?set([goal(perform(action1)), goal(perform(action2))]))")])
        )

    def test_findout_for_yn_question_for_goal_proposition(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <predicate name="price" sort="real"/>
</ontology>"""
        )

        self._when_compile_domain_with_plan(
            """
<findout type="yn_question">
  <resolve type="wh_question" predicate="price"/>
</findout>"""
        )

        self._then_result_has_plan(Plan([self._parse("findout(?goal(resolve(?X.price(X))))")]))

    def test_findout_for_yn_question_for_goal_proposition_perform(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <action name="apply_for_membership"/>
</ontology>"""
        )

        self._when_compile_domain_with_plan(
            """
<findout type="yn_question">
  <perform action="apply_for_membership"/>
</findout>"""
        )

        self._then_result_has_plan(Plan([self._parse("findout(?goal(perform(apply_for_membership)))")]))

    def test_if_then_else(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <sort name="how"/>
  <sort name="train_type"/>
  <predicate name="means_of_transport" sort="how"/>
  <predicate name="transport_train_type" sort="train_type"/>
  <individual name="train" sort="how"/>
</ontology>"""
        )

        self._when_compile_domain_with_plan(
            """
<if>
  <condition><proposition predicate="means_of_transport" value="train"/></condition>
  <then>
    <findout type="wh_question" predicate="transport_train_type" />
  </then>
  <else />
</if>"""
        )

        self._then_result_has_plan(
            Plan([
                IfThenElse(
                    self._parse("means_of_transport(train)"), self._parse("findout(?X.transport_train_type(X))"), None
                )
            ])
        )

    def test_forget_proposition(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <sort name="how"/>
  <predicate name="means_of_transport" sort="how"/>
  <individual name="train" sort="how"/>
</ontology>"""
        )

        self._when_compile_domain_with_plan(
            '<forget><proposition predicate="means_of_transport" value="train"/></forget>'
        )

        self._then_result_has_plan(Plan([self._parse("forget(means_of_transport(train))")]))

    def test_forget_predicate(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <sort name="how"/>
  <predicate name="means_of_transport" sort="how"/>
</ontology>"""
        )

        self._when_compile_domain_with_plan('<forget predicate="means_of_transport"/>')

        self._then_result_has_plan(Plan([self._parse("forget(means_of_transport)")]))

    def test_forget_all(self):
        self._given_compiled_ontology()
        self._when_compile_domain_with_plan('<forget_all />')
        self._then_result_has_plan(Plan([ForgetAllPlanItem()]))

    def test_invoke_service_query(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <predicate name="price" sort="real"/>
</ontology>"""
        )
        self._when_compile_domain_with_plan('<invoke_service_query type="wh_question" predicate="price"/>')
        self._then_result_has_plan(
            Plan([InvokeServiceQueryPlanItem(self._parse("?X.price(X)"), min_results=1, max_results=1)])
        )

    def test_invoke_service_action_with_device_attribute(self):
        self._given_compiled_ontology()
        self._when_compile_domain_with_plan_then_exception_is_raised_matching(
            '<invoke_service_action device="MockupDevice" name="MockupAction" />', UnexpectedAttributeException,
            "Attribute 'device=\"MockupDevice\"' is not supported in <invoke_service_action name='MockupAction'>. "
            "Use 'service_interface.xml' instead."
        )

    def test_invoke_service_query_with_device_attribute(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <predicate name="dest_city" sort="city"/>
  <sort name="city"/>
</ontology>"""
        )
        self._when_compile_domain_with_plan_then_exception_is_raised_matching(
            '<invoke_service_query type="wh_question" predicate="dest_city" device="MockupDevice"/>',
            UnexpectedAttributeException, "Attribute 'device=\"MockupDevice\"' is not supported in "
            "<invoke_service_query ... predicate='dest_city'>. Use 'service_interface.xml' instead."
        )

    def _when_compile_domain_with_plan_then_exception_is_raised_matching(
        self, plan_xml, expected_exception, expected_message
    ):
        with self.assertRaisesRegexp(expected_exception, expected_message):
            self._when_compile_domain_with_plan(plan_xml)

    def test_dev_query_deprecation_warning(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <predicate name="price" sort="real"/>
</ontology>"""
        )
        self._when_compile_domain_with_plan_then_deprecation_warning_is_issued(
            '<dev_query type="wh_question" predicate="price" />',
            '<dev_query> is deprecated. Use <invoke_service_query> instead.'
        )

    def test_deprecated_dev_query(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <predicate name="price" sort="real"/>
</ontology>"""
        )
        self._when_compile_domain_with_plan_ignoring_warnings('<dev_query type="wh_question" predicate="price" />')
        self._then_result_has_plan(
            Plan([InvokeServiceQueryPlanItem(self._parse("?X.price(X)"), min_results=1, max_results=1)])
        )

    def _when_compile_domain_with_plan_and_ignoring_warnings_then_exception_is_raised_matching(self, *args, **kwargs):
        with self._mock_warnings():
            self._when_compile_domain_with_plan_then_exception_is_raised_matching(*args, **kwargs)

    def test_invoke_service_action_default_attributes(self):
        self._given_compiled_ontology()
        self._when_compile_domain_with_plan('<invoke_service_action name="MockupAction" />')
        self._then_result_has_plan(Plan([InvokeServiceActionPlanItem("MockupOntology", "MockupAction")]))

    def test_invoke_service_action_override_preconfirm(self):
        self._given_compiled_ontology()
        self._when_compile_domain_with_plan(
            """
          <invoke_service_action name="MockupAction" preconfirm="assertive" />"""
        )
        self._then_result_has_plan(
            Plan([
                InvokeServiceActionPlanItem(
                    u"MockupOntology", u"MockupAction", preconfirm=InvokeServiceActionPlanItem.ASSERTIVE
                )
            ])
        )

    def test_invoke_service_action_override_postconfirm(self):
        self._given_compiled_ontology()
        self._when_compile_domain_with_plan('<invoke_service_action name="MockupAction" postconfirm="true" />')
        self._then_result_has_plan(
            Plan([InvokeServiceActionPlanItem("MockupOntology", "MockupAction", postconfirm=True)])
        )

    def test_invoke_service_action_override_downdate_plan(self):
        self._given_compiled_ontology()
        self._when_compile_domain_with_plan(
            """
          <invoke_service_action name="MockupAction" downdate_plan="false" />"""
        )
        self._then_result_has_plan(
            Plan([InvokeServiceActionPlanItem("MockupOntology", "MockupAction", downdate_plan=False)])
        )

    def test_deprecated_dev_perform_default_attributes(self):
        self._given_compiled_ontology()
        self._when_compile_domain_with_plan_ignoring_warnings('<dev_perform action="MockupAction" />')
        self._then_result_has_plan(Plan([InvokeServiceActionPlanItem("MockupOntology", "MockupAction")]))

    def _when_compile_domain_with_plan_ignoring_warnings(self, plan_xml):
        with self._mock_warnings():
            self._when_compile_domain_with_plan(plan_xml)

    def test_deprecated_dev_perform_override_preconfirm(self):
        self._given_compiled_ontology()
        self._when_compile_domain_with_plan_ignoring_warnings(
            """
          <dev_perform action="MockupAction" preconfirm="assertive" />"""
        )
        self._then_result_has_plan(
            Plan([
                InvokeServiceActionPlanItem(
                    "MockupOntology", "MockupAction", preconfirm=InvokeServiceActionPlanItem.ASSERTIVE
                )
            ])
        )

    def test_deprecated_dev_perform_override_postconfirm(self):
        self._given_compiled_ontology()
        self._when_compile_domain_with_plan_ignoring_warnings(
            '<dev_perform action="MockupAction" postconfirm="true" />'
        )
        self._then_result_has_plan(
            Plan([InvokeServiceActionPlanItem("MockupOntology", "MockupAction", postconfirm=True)])
        )

    def test_dev_perform_deprecation_warning(self):
        self._given_compiled_ontology()
        self._when_compile_domain_with_plan_then_deprecation_warning_is_issued(
            '<dev_perform action="MockupAction" />', "<dev_perform> is deprecated. Use <invoke_service_action> instead."
        )

    def _when_compile_domain_with_plan_then_deprecation_warning_is_issued(self, plan_xml, expected_warning_message):
        with self._mock_warnings() as mocked_warnings:
            self._when_compile_domain_with_plan(plan_xml)
            self._then_warning_is_issued(mocked_warnings, expected_warning_message)

    def test_jumpto(self):
        self._given_compiled_ontology()
        self._when_compile_domain_with_plan('<jumpto type="perform" action="top"/>')
        self._then_result_has_plan(Plan([self._parser.parse("jumpto(perform(top))")]))

    def test_assume_shared(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <predicate name="price" sort="real"/>
</ontology>"""
        )
        self._when_compile_domain_with_plan(
            """
<assume_shared>
  <proposition predicate="price" value="123.0"/>
</assume_shared>"""
        )
        self._then_result_has_plan(Plan([self._parser.parse("assume_shared(price(123.0))")]))


class GrammarCompilerTests(DddXmlCompilerTestCase):
    def test_multiple_variants(self):
        self._given_compiled_ontology("""
<ontology name="Ontology">
  <action name="buy"/>
</ontology>""")

        self._when_compile_grammar(
            """
<grammar>
  <action name="buy">
    <one-of>
      <item>buy</item>
      <item>purchase</item>
    </one-of>
  </action>
</grammar>
"""
        )

        self._then_result_is(
            Node(
                Constants.GRAMMAR, {}, [
                    Node(
                        Constants.ACTION, {"name": "buy"}, [
                            Node(
                                Constants.ONE_OF, {},
                                [Node(Constants.ITEM, {}, ["buy"]),
                                 Node(Constants.ITEM, {}, ["purchase"])]
                            )
                        ]
                    )
                ]
            )
        )

    def test_action(self):
        self._given_compiled_ontology("""
<ontology name="Ontology">
  <action name="buy"/>
</ontology>""")

        self._when_compile_grammar("""
<grammar>
  <action name="buy">purchase</action>
</grammar>
""")

        self._then_grammar_is([Node(Constants.ACTION, {"name": "buy"}, [Node(Constants.ITEM, {}, ["purchase"])])])

    def test_predicate_explicit_speaker(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <predicate name="price" sort="real"/>
</ontology>"""
        )

        self._when_compile_grammar(
            """
<grammar>
  <question predicate="price" speaker="all">price information</question>
</grammar>"""
        )

        self._then_grammar_is([
            Node(Constants.PREDICATE, {"name": "price"}, [Node(Constants.ITEM, {}, ["price information"])])
        ])

    def test_predicate_implicit_speaker(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <predicate name="price" sort="real"/>
</ontology>"""
        )

        self._when_compile_grammar(
            """
<grammar>
  <question predicate="price">price information</question>
</grammar>"""
        )

        self._then_grammar_is([
            Node(Constants.PREDICATE, {"name": "price"}, [Node(Constants.ITEM, {}, ["price information"])])
        ])

    def test_user_question(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <sort name="city"/>
  <predicate name="price" sort="real"/>
</ontology>"""
        )

        self._when_compile_grammar(
            """
<grammar>
  <question predicate="price" speaker="user">price for travelling to <slot sort="city"/>
  </question>
</grammar>"""
        )

        self._then_grammar_is([
            Node(
                Constants.USER_QUESTION, {"predicate": "price"},
                [Node(Constants.ITEM, {},
                      ["price for travelling to ", Node(Constants.SLOT, {"sort": "city"})])]
            )
        ])

    def test_system_question(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <sort name="city"/>
  <predicate name="dest_city" sort="city"/>
</ontology>"""
        )

        self._when_compile_grammar(
            """
<grammar>
  <question predicate="dest_city" speaker="system">where do you want to go</question>
</grammar>"""
        )

        self._then_grammar_is([
            Node(
                Constants.SYS_QUESTION, {"predicate": "dest_city"},
                [Node(Constants.ITEM, {}, ["where do you want to go"])]
            )
        ])

    def test_english_noun_phrase_as_single_variant(self):
        self._given_compiled_ontology()
        self._when_compile_grammar(
            """
<grammar>
  <action name="top">
    <np>
      <indefinite>start view</indefinite>
      <definite>the start view</definite>
    </np>
  </action>
</grammar>"""
        )
        self._then_grammar_is([
            Node(
                Constants.ACTION, {"name": "top"}, [
                    Node(
                        Constants.NP, {}, [
                            Node(Constants.INDEFINITE, {}, ["start view"]),
                            Node(Constants.DEFINITE, {}, ["the start view"])
                        ]
                    )
                ]
            )
        ])

    def test_english_noun_phrase_among_multiple_variants(self):
        self._given_compiled_ontology()
        self._when_compile_grammar(
            """
<grammar>
  <action name="top">
    <one-of>
      <item>
        <np>
          <indefinite>start view</indefinite>
          <definite>the start view</definite>
        </np>
      </item>
      <item>main menu</item>
    </one-of>
  </action>
</grammar>"""
        )
        self._then_grammar_is([
            Node(
                Constants.ACTION, {"name": "top"}, [
                    Node(
                        Constants.ONE_OF, {}, [
                            Node(
                                Constants.ITEM, {}, [
                                    Node(
                                        Constants.NP, {}, [
                                            Node(Constants.INDEFINITE, {}, ["start view"]),
                                            Node(Constants.DEFINITE, {}, ["the start view"])
                                        ]
                                    )
                                ]
                            ),
                            Node(Constants.ITEM, {}, ["main menu"])
                        ]
                    )
                ]
            )
        ])

    def test_english_verb_phrase_as_single_variant(self):
        self._given_compiled_ontology("""
<ontology name="Ontology">
  <action name="buy"/>
</ontology>""")
        self._when_compile_grammar(
            """
<grammar>
  <action name="buy">
    <vp>
      <infinitive>buy</infinitive>
      <imperative>buy</imperative>
      <ing-form>buying</ing-form>
      <object>a ticket</object>
    </vp>
  </action>
</grammar>"""
        )
        self._then_grammar_is([
            Node(
                Constants.ACTION, {"name": "buy"}, [
                    Node(
                        Constants.VP, {}, [
                            Node(Constants.INFINITIVE, {}, ["buy"]),
                            Node(Constants.IMPERATIVE, {}, ["buy"]),
                            Node(Constants.ING_FORM, {}, ["buying"]),
                            Node(Constants.OBJECT, {}, ["a ticket"])
                        ]
                    )
                ]
            )
        ])

    def test_english_verb_phrase_among_multiple_variants(self):
        self._given_compiled_ontology("""
<ontology name="Ontology">
  <action name="buy"/>
</ontology>""")
        self._when_compile_grammar(
            """
<grammar>
  <action name="buy">
    <one-of>
      <item>
        <vp>
          <infinitive>buy</infinitive>
          <imperative>buy</imperative>
          <ing-form>buying</ing-form>
          <object>a ticket</object>
        </vp>
      </item>
      <item>purchase</item>
    </one-of>
  </action>
</grammar>"""
        )
        self._then_grammar_is([
            Node(
                Constants.ACTION, {"name": "buy"}, [
                    Node(
                        Constants.ONE_OF, {}, [
                            Node(
                                Constants.ITEM, {}, [
                                    Node(
                                        Constants.VP, {}, [
                                            Node(Constants.INFINITIVE, {}, ["buy"]),
                                            Node(Constants.IMPERATIVE, {}, ["buy"]),
                                            Node(Constants.ING_FORM, {}, ["buying"]),
                                            Node(Constants.OBJECT, {}, ["a ticket"])
                                        ]
                                    )
                                ]
                            ),
                            Node(Constants.ITEM, {}, ["purchase"])
                        ]
                    )
                ]
            )
        ])

    def test_swedish_verb_phrase(self):
        self._given_compiled_ontology("""
<ontology name="Ontology">
  <action name="buy"/>
</ontology>""")
        self._when_compile_grammar(
            """
<grammar>
  <action name="buy">
    <vp>
      <infinitive>köpa</infinitive>
      <imperative>köp</imperative>
      <ing-form>köper</ing-form>
      <object>en biljett</object>
    </vp>
  </action>
</grammar>"""
        )
        self._then_grammar_is([
            Node(
                Constants.ACTION, {"name": "buy"}, [
                    Node(
                        Constants.VP, {}, [
                            Node(Constants.INFINITIVE, {}, [u"köpa"]),
                            Node(Constants.IMPERATIVE, {}, [u"köp"]),
                            Node(Constants.ING_FORM, {}, [u"köper"]),
                            Node(Constants.OBJECT, {}, [u"en biljett"])
                        ]
                    )
                ]
            )
        ])

    def test_spanish_verb_phrase(self):
        self._given_compiled_ontology("""
    <ontology name="Ontology">
      <action name="buy"/>
    </ontology>""")
        self._when_compile_grammar(
            """
    <grammar>
      <action name="buy">
        <vp>
          <infinitive>comprar</infinitive>
          <imperative>compra</imperative>
          <ing-form>comprando</ing-form>
          <object>un billete</object>
        </vp>
      </action>
    </grammar>"""
        )
        self._then_grammar_is([
            Node(
                Constants.ACTION, {"name": "buy"}, [
                    Node(
                        Constants.VP, {}, [
                            Node(Constants.INFINITIVE, {}, ["comprar"]),
                            Node(Constants.IMPERATIVE, {}, ["compra"]),
                            Node(Constants.ING_FORM, {}, ["comprando"]),
                            Node(Constants.OBJECT, {}, ["un billete"])
                        ]
                    )
                ]
            )
        ])

    def test_answer_combination(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <sort name="city"/>
  <predicate name="dept_city" sort="city"/>
  <predicate name="dest_city" sort="city"/>
</ontology>"""
        )

        self._when_compile_grammar(
            """
<grammar>
  <answer speaker="user">
    <one-of>
      <item>
        from <slot type="individual" predicate="dept_city" /> to <slot type="individual" predicate="dest_city" />
      </item>
      <item>
        from <slot type="individual" predicate="dept_city" />
      </item>
    </one-of>
  </answer>
</grammar>"""
        )

        self._then_grammar_is([
            Node(
                Constants.ANSWER_COMBINATION, {}, [
                    Node(
                        Constants.ONE_OF, {}, [
                            Node(
                                Constants.ITEM, {}, [
                                    "from ",
                                    Node(Constants.SLOT, {"predicate": "dept_city"}), " to ",
                                    Node(Constants.SLOT, {"predicate": "dest_city"})
                                ]
                            ),
                            Node(Constants.ITEM, {},
                                 ["from ", Node(Constants.SLOT, {"predicate": "dept_city"})])
                        ]
                    )
                ]
            )
        ])

    def test_report_started(self):
        self._given_compiled_ontology()
        self._given_service_interface_with_action("IncomingCall")
        self._when_compile_grammar(
            """
<grammar>
  <report action="IncomingCall" status="started">incoming call</report>
</grammar>"""
        )
        self._then_grammar_is([
            Node(Constants.REPORT_STARTED, {"action": "IncomingCall"}, [Node(Constants.ITEM, {}, ["incoming call"])])
        ])

    def test_report_ended(self):
        self._given_compiled_ontology()
        self._given_service_interface_with_action("IncomingCall")
        self._when_compile_grammar(
            """
<grammar>
  <report action="IncomingCall" status="ended">call ended</report>
</grammar>"""
        )
        self._then_grammar_is([
            Node(Constants.REPORT_ENDED, {"action": "IncomingCall"}, [Node(Constants.ITEM, {}, ["call ended"])])
        ])

    def test_action_with_empty_string(self):
        self._given_compiled_ontology()
        self._given_service_interface_with_action("IncomingCall")
        self._when_compile_grammar(
            """
<grammar>
  <report action="IncomingCall" status="ended">
  </report>
</grammar>"""
        )
        self._then_grammar_is([Node(Constants.REPORT_ENDED, {"action": "IncomingCall"}, [None])])

    def test_report_failed(self):
        self._given_compiled_ontology()
        self._given_service_interface_with_action("CancelReservation")
        self._when_compile_grammar(
            """
<grammar>
  <report action="CancelReservation" status="failed" reason="no_reservation_exists">there is no reservation</report>
</grammar>"""
        )
        self._then_grammar_is([
            Node(
                Constants.REPORT_FAILED, {
                    "action": "CancelReservation",
                    "reason": "no_reservation_exists"
                }, [Node(Constants.ITEM, {}, ["there is no reservation"])]
            )
        ])

    def test_individual(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <sort name="city"/>
  <individual name="city1" sort="city"/>
</ontology>"""
        )
        self._when_compile_grammar("""
<grammar>
  <individual name="city_1">paris</individual>
</grammar>""")
        self._then_grammar_is([Node(Constants.INDIVIDUAL, {"name": "city_1"}, [Node(Constants.ITEM, {}, ["paris"])])])

    def test_system_answer(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <sort name="city"/>
  <predicate name="dest_city" sort="city"/>
</ontology>"""
        )
        self._when_compile_grammar(
            """
<grammar>
  <answer speaker="system">to <slot type="individual" predicate="dest_city"/></answer>
</grammar>"""
        )
        self._then_grammar_is([
            Node(
                Constants.SYS_ANSWER, {"predicate": "dest_city"},
                [Node(Constants.ITEM, {}, ["to ", Node(Constants.SLOT, {})])]
            )
        ])

    def test_system_answer_with_embedded_answer(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <sort name="city"/>
  <predicate name="dest_city" sort="city"/>
  <predicate name="price" sort="real"/>
</ontology>"""
        )
        self._when_compile_grammar(
            """
<grammar>
  <answer speaker="system" predicate="price">
    the price to <slot type="individual" predicate="dest_city"/> is <slot type="individual" predicate="price"/>
  </answer>
</grammar>"""
        )
        self._then_grammar_is([
            Node(
                Constants.SYS_ANSWER, {"predicate": "price"}, [
                    Node(
                        Constants.ITEM, {}, [
                            "the price to ",
                            Node(Constants.SLOT, {"predicate": "dest_city"}), " is ",
                            Node(Constants.SLOT, {})
                        ]
                    )
                ]
            )
        ])

    def test_positive_system_answer(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <predicate name="qualified" sort="boolean"/>
</ontology>"""
        )
        self._when_compile_grammar(
            """
<grammar>
  <answer speaker="system" predicate="qualified" polarity="positive">you are qualified</answer>
</grammar>"""
        )
        self._then_grammar_is([
            Node(
                Constants.POSITIVE_SYS_ANSWER, {"predicate": "qualified"},
                [Node(Constants.ITEM, {}, ["you are qualified"])]
            )
        ])

    def test_negative_system_answer(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <predicate name="qualified" sort="boolean"/>
</ontology>"""
        )
        self._when_compile_grammar(
            """
<grammar>
  <answer speaker="system" predicate="qualified" polarity="negative">you are not qualified</answer>
</grammar>"""
        )
        self._then_grammar_is([
            Node(
                Constants.NEGATIVE_SYS_ANSWER, {"predicate": "qualified"},
                [Node(Constants.ITEM, {}, ["you are not qualified"])]
            )
        ])

    def test_action_title(self):
        self._given_compiled_ontology()
        self._when_compile_grammar("""
<grammar>
  <title type="action" action="top">main menu</title>
</grammar>""")
        self._then_grammar_is([
            Node(Constants.ACTION_TITLE, {"action": "top"}, [Node(Constants.ITEM, {}, ["main menu"])])
        ])

    def test_issue_title(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <predicate name="price" sort="real"/>
</ontology>"""
        )
        self._when_compile_grammar(
            """
<grammar>
  <title type="question" predicate="price">price information</title>
</grammar>"""
        )
        self._then_grammar_is([
            Node(Constants.ISSUE_TITLE, {"predicate": "price"}, [Node(Constants.ITEM, {}, ["price information"])])
        ])

    def test_validity(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <sort name="city"/>
  <predicate name="dest_city" sort="city"/>
</ontology>"""
        )
        self._given_service_interface_with_validity("CityValidity")
        self._when_compile_grammar(
            """
<grammar>
  <validity name="CityValidity">
    invalid city <slot type="individual" predicate="dest_city" />
  </validity>
</grammar>"""
        )
        self._then_grammar_is([
            Node(
                Constants.VALIDITY, {"name": "CityValidity"},
                [Node(Constants.ITEM, {},
                      ["invalid city ", Node(Constants.SLOT, {"predicate": "dest_city"})])]
            )
        ])

    def test_prereport(self):
        self._given_compiled_ontology()
        self._given_service_interface_with_action("MakeReservation")
        self._when_compile_grammar(
            """
<grammar>
  <report action="MakeReservation" status="started" source="dialogue">making the reservation</report>
</grammar>"""
        )
        self._then_grammar_is([
            Node(
                Constants.PREREPORT, {"action": "MakeReservation"},
                [Node(Constants.ITEM, {}, ["making the reservation"])]
            )
        ])

    def test_preconfirm(self):
        self._given_compiled_ontology()
        self._given_service_interface_with_action("MakeReservation")
        self._when_compile_grammar(
            """
<grammar>
  <preconfirm action="MakeReservation">would you like to make the reservation</preconfirm>
</grammar>"""
        )
        self._then_grammar_is([
            Node(
                Constants.PRECONFIRM, {"action": "MakeReservation"},
                [Node(Constants.ITEM, {}, ["would you like to make the reservation"])]
            )
        ])

    def test_greeting(self):
        self._given_compiled_ontology()
        self._when_compile_grammar("""
<grammar>
  <greeting>Welcome</greeting>
</grammar>""")
        self._then_grammar_is([Node(Constants.GREETING, {}, [Node(Constants.ITEM, {}, ["Welcome"])])])

    def test_whitespacing_for_trailing_text_node(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <sort name="city"/>
  <predicate name="dest_city" sort="city"/>
</ontology>"""
        )
        self._when_compile_grammar(
            """
<grammar>
  <answer speaker="system">
    <slot type="individual" predicate="dest_city"/> as destination
  </answer>
</grammar>"""
        )
        self._then_grammar_is([
            Node(
                Constants.SYS_ANSWER, {"predicate": "dest_city"},
                [Node(Constants.ITEM, {}, [Node(Constants.SLOT, {}), " as destination"])]
            )
        ])

    def test_whitespacing_between_slots(self):
        self._given_compiled_ontology(
            """
<ontology name="Ontology">
  <predicate name="alarm_hour" sort="integer"/>
  <predicate name="alarm_minute" sort="integer"/>
</ontology>"""
        )
        self._when_compile_grammar(
            """
<grammar>
  <report action="AlarmRings" status="started">
    the time is <slot type="individual" predicate="alarm_hour"/> <slot type="individual" predicate="alarm_minute"/>
  </report>
</grammar>"""
        )
        self._then_grammar_is([
            Node(
                Constants.REPORT_STARTED, {"action": "AlarmRings"}, [
                    Node(
                        Constants.ITEM, {}, [
                            "the time is ",
                            Node(Constants.SLOT, {"predicate": "alarm_hour"}), " ",
                            Node(Constants.SLOT, {"predicate": "alarm_minute"})
                        ]
                    )
                ]
            )
        ])

    def test_unexpected_element_yields_exception(self):
        self._given_compiled_ontology()
        with self.assertRaises(ViolatesSchemaException):
            self._when_compile_grammar("""
<grammar>
  <unexpected_element/>
</grammar>""")

    def test_tolerate_comment(self):
        self._given_compiled_ontology("""
<ontology name="Ontology">
  <action name="buy"/>
</ontology>""")

        self._when_compile_grammar(
            """
<grammar>
  <action name="buy">purchase</action>
  <!-- this is a comment -->
</grammar>
"""
        )

        self._then_grammar_is([Node(Constants.ACTION, {"name": "buy"}, [Node(Constants.ITEM, {}, ["purchase"])])])

    def setUp(self):
        super(GrammarCompilerTests, self).setUp()
        self._device_handler = None


class RglGrammarCompilerTests(DddXmlCompilerTestCase):
    def test_action_with_noun_phrase(self):
        self._given_compiled_ontology()

        self._when_compile_rgl_grammar(
            """
<grammar>
  <action name="top">
    <noun-phrase>
      <noun ref="menu"/>
    </noun-phrase>
  </action>
</grammar>"""
        )

        self._then_grammar_is([
            Node(
                Constants.ACTION, {"name": "top"},
                [Node(rgl_types.NOUN_PHRASE, {}, [Node(rgl_types.NOUN, {"ref": "menu"})])]
            )
        ])

    def test_lexicon_with_noun(self):
        self._given_compiled_ontology()

        self._when_compile_rgl_grammar(
            """
<grammar>
  <lexicon>
    <noun id="menu">
      <singular>menu</singular>
    </noun>
  </lexicon>
</grammar>"""
        )

        self._then_grammar_is([
            Node(rgl_types.LEXICON, {}, [Node(rgl_types.NOUN, {"id": "menu"}, [Node("singular", {}, ["menu"])])])
        ])

    def test_action_with_verb_phrase(self):
        self._given_compiled_ontology()

        self._when_compile_rgl_grammar(
            """
<grammar>
  <action name="top">
    <verb-phrase>
      <verb ref="restart"/>
    </verb-phrase>
  </action>
</grammar>"""
        )

        self._then_grammar_is([
            Node(
                Constants.ACTION, {"name": "top"},
                [Node(rgl_types.VERB_PHRASE, {}, [Node(rgl_types.VERB, {"ref": "restart"})])]
            )
        ])

    def test_predicate_with_noun_phrase(self):
        self._given_compiled_ontology()

        self._when_compile_rgl_grammar(
            """
<grammar>
  <predicate name="phone_number_of_contact">
    <noun-phrase>
      <noun ref="number" />
    </noun-phrase>
  </predicate>
</grammar>"""
        )

        self._then_grammar_is([
            Node(
                Constants.PREDICATE, {"name": "phone_number_of_contact"},
                [Node(rgl_types.NOUN_PHRASE, {}, [Node(rgl_types.NOUN, {"ref": "number"})])]
            )
        ])

    def test_lexicon_with_verb(self):
        self._given_compiled_ontology()

        self._when_compile_rgl_grammar(
            """
<grammar>
  <lexicon>
    <verb id="restart">
      <infinitive>restart</infinitive>
    </verb>
  </lexicon>
</grammar>"""
        )

        self._then_grammar_is([
            Node(
                rgl_types.LEXICON, {}, [Node(rgl_types.VERB, {"id": "restart"}, [Node("infinitive", {}, ["restart"])])]
            )
        ])

    def test_one_of_for_action(self):
        self._given_compiled_ontology()

        self._when_compile_rgl_grammar(
            """
<grammar>
  <action name="top">
    <one-of>
      <item>
        <verb-phrase>
          <verb ref="restart"/>
        </verb-phrase>
      </item>
      <item>
        <verb-phrase>
          <verb ref="forget"/>
        </verb-phrase>
      </item>
    </one-of>
  </action>
</grammar>"""
        )

        self._then_grammar_is([
            Node(
                Constants.ACTION, {"name": "top"}, [
                    Node(
                        Constants.ONE_OF, {}, [
                            Node(
                                Constants.ITEM, {},
                                [Node(rgl_types.VERB_PHRASE, {}, [Node(rgl_types.VERB, {"ref": "restart"})])]
                            ),
                            Node(
                                Constants.ITEM, {},
                                [Node(rgl_types.VERB_PHRASE, {}, [Node(rgl_types.VERB, {"ref": "forget"})])]
                            )
                        ]
                    )
                ]
            )
        ])

    def test_simple_request(self):
        self._given_compiled_ontology()

        self._when_compile_rgl_grammar(
            """
<grammar>
  <request action="top"><utterance>forget everything</utterance></request>
</grammar>
"""
        )

        self._then_grammar_is([
            Node(rgl_types.REQUEST, {"action": "top"}, [Node(rgl_types.UTTERANCE, {}, ["forget everything"])])
        ])

    def test_request_with_individual(self):
        self._given_compiled_ontology("""
<ontology>
  <action name="call"/>
  <sort name="contact"/>
</ontology>
""")

        self._when_compile_rgl_grammar(
            """
<grammar>
  <request action="call">
    <utterance>call <individual sort="contact"/></utterance>
  </request>
</grammar>
"""
        )

        self._then_grammar_is([
            Node(
                rgl_types.REQUEST, {"action": "call"},
                [Node(rgl_types.UTTERANCE, {}, ["call ", Node(Constants.INDIVIDUAL, {"sort": "contact"}, [])])]
            )
        ])

    def test_individual_as_proper_noun(self):
        self._given_compiled_ontology(
            """
<ontology>
  <sort name="city"/>
  <individual name="city1" sort="city"/>
</ontology>"""
        )
        self._when_compile_rgl_grammar(
            """
<grammar>
  <individual name="city_1">
    <proper-noun>Paris</proper-noun>
  </individual>
</grammar>"""
        )
        self._then_grammar_is([
            Node(Constants.INDIVIDUAL, {"name": "city_1"}, [Node(rgl_types.PROPER_NOUN, {}, ["Paris"])])
        ])

    def test_report_ended(self):
        self._given_compiled_ontology()
        self._given_service_interface_with_action("IncomingCall")
        self._when_compile_rgl_grammar(
            """
<grammar>
  <report action="IncomingCall" status="ended"><utterance>call ended</utterance></report>
</grammar>"""
        )
        self._then_grammar_is([
            Node(Constants.REPORT_ENDED, {"action": "IncomingCall"}, [Node(rgl_types.UTTERANCE, {}, ["call ended"])])
        ])

    def test_system_question(self):
        self._given_compiled_ontology(
            """
<ontology>
  <sort name="city"/>
  <predicate name="dest_city" sort="city"/>
</ontology>"""
        )

        self._when_compile_rgl_grammar(
            """
<grammar>
  <question predicate="dest_city" speaker="system">
    <utterance>where do you want to go</utterance>
  </question>
</grammar>"""
        )

        self._then_grammar_is([
            Node(
                Constants.SYS_QUESTION, {"predicate": "dest_city"},
                [Node(rgl_types.UTTERANCE, {}, ["where do you want to go"])]
            )
        ])


class ServiceInterfaceCompilerTests(DddXmlCompilerTestCase):
    def test_action_without_parameters(self):
        self._when_compile_service_interface(
            """
<service_interface>
  <action name="EndCall">
    <parameters/>
    <failure_reasons/>
    <target>
      <device_module device="CallDevice"/>
    </target>
  </action>
</service_interface>
"""
        )
        self._then_service_interface_has_actions([
            ServiceActionInterface(
                "EndCall", target=DeviceModuleTarget("CallDevice"), parameters=[], failure_reasons=[]
            )
        ])

    def _when_compile_service_interface(self, device_xml):
        self._result = DddXmlCompiler().compile_service_interface(device_xml)

    def _then_service_interface_has_actions(self, expected_actions):
        self.assertEqual(self._result.actions, expected_actions)

    def test_action_with_parameters(self):
        self._when_compile_service_interface(
            """
<service_interface>
  <action name="CallContact">
    <parameters>
      <parameter predicate="contact_to_call"/>
      <parameter predicate="number_to_call"/>
    </parameters>
    <failure_reasons/>
    <target>
      <device_module device="CallDevice"/>
    </target>
  </action>
</service_interface>
"""
        )
        self._then_service_interface_has_actions([
            ServiceActionInterface(
                "CallContact",
                target=DeviceModuleTarget("CallDevice"),
                parameters=[
                    ServiceParameter("contact_to_call", ParameterField.VALUE, is_optional=False),
                    ServiceParameter("number_to_call", ParameterField.VALUE, is_optional=False),
                ],
                failure_reasons=[]
            )
        ])

    def test_action_with_failure_reasons(self):
        self._when_compile_service_interface(
            """
<service_interface>
  <action name="CallContact">
    <parameters>
      <parameter predicate="contact_to_call"/>
      <parameter predicate="number_to_call"/>
    </parameters>
    <failure_reasons>
      <failure_reason name="invalid_number"/>
      <failure_reason name="missing_number"/>
    </failure_reasons>
    <target>
      <device_module device="CallDevice"/>
    </target>
  </action>
</service_interface>
"""
        )
        self._then_service_interface_has_actions([
            ServiceActionInterface(
                "CallContact",
                target=DeviceModuleTarget("CallDevice"),
                parameters=[
                    ServiceParameter("contact_to_call", ParameterField.VALUE, is_optional=False),
                    ServiceParameter("number_to_call", ParameterField.VALUE, is_optional=False),
                ],
                failure_reasons=[
                    ActionFailureReason("invalid_number"),
                    ActionFailureReason("missing_number"),
                ]
            )
        ])

    def test_optional_parameter(self):
        self._when_compile_service_interface(
            """
<service_interface>
  <action name="Snooze">
    <parameters>
      <parameter predicate="minutes_to_snooze" optional="true"/>
    </parameters>
    <failure_reasons/>
    <target>
      <device_module device="CallDevice"/>
    </target>
  </action>
</service_interface>
"""
        )
        self._then_service_interface_has_actions([
            ServiceActionInterface(
                "Snooze",
                target=DeviceModuleTarget("CallDevice"),
                parameters=[
                    ServiceParameter("minutes_to_snooze", ParameterField.VALUE, is_optional=True),
                ],
                failure_reasons=[]
            )
        ])

    def test_parameter_field(self):
        self._when_compile_service_interface(
            """
<service_interface>
  <action name="CallContact">
    <parameters>
      <parameter predicate="contact_to_call" format="grammar_entry"/>
    </parameters>
    <failure_reasons/>
    <target>
      <device_module device="CallDevice"/>
    </target>
  </action>
</service_interface>
"""
        )
        self._then_service_interface_has_actions([
            ServiceActionInterface(
                "CallContact",
                target=DeviceModuleTarget("CallDevice"),
                parameters=[
                    ServiceParameter("contact_to_call", ParameterField.GRAMMAR_ENTRY),
                ],
                failure_reasons=[]
            )
        ])

    def test_unexpected_parameter_field(self):
        self._when_compile_service_interface_then_exception_is_raised_matching(
            """
<service_interface>
  <action name="CallContact">
    <parameters>
      <parameter predicate="contact_to_call" format="something_else"/>
    </parameters>
    <failure_reasons/>
    <target>
      <device_module device="CallDevice"/>
    </target>
  </action>
</service_interface>
""", ViolatesSchemaException,
            "Expected service_interface.xml compliant with schema but it's in violation: Element 'parameter', "
            "attribute 'format': \[facet 'enumeration'\] The value 'something_else' is not an element of the set "
            "\{'value', 'grammar_entry'\}., line 5"
        )

    def _when_compile_service_interface_then_exception_is_raised_matching(
        self, xml, expected_exception, expected_message
    ):

        with self.assertRaisesRegexp(expected_exception, expected_message):
            self._when_compile_service_interface(xml)

    def test_query_with_device_module_target(self):
        self._when_compile_service_interface(
            """
<service_interface>
  <query name="available_contact">
    <parameters>
      <parameter predicate="contact_to_call"/>
    </parameters>
    <target>
      <device_module device="CallDevice"/>
    </target>
  </query>
</service_interface>
"""
        )
        self._then_service_interface_has_queries([
            ServiceQueryInterface(
                "available_contact",
                target=DeviceModuleTarget("CallDevice"),
                parameters=[
                    ServiceParameter("contact_to_call", ParameterField.VALUE),
                ]
            )
        ])

    def _then_service_interface_has_queries(self, expected_queries):
        self.assertEqual(self._result.queries, expected_queries)

    def test_entity_recognizer_with_device_module_target(self):
        self._when_compile_service_interface(
            """
<service_interface>
  <entity_recognizer name="ContactNameRecognizer">
    <target>
      <device_module device="CallDevice"/>
    </target>
  </entity_recognizer>
</service_interface>
"""
        )
        self._then_service_interface_has_entity_recognizers([
            ServiceEntityRecognizerInterface("ContactNameRecognizer", target=DeviceModuleTarget("CallDevice")),
        ])

    def _then_service_interface_has_entity_recognizers(self, expected_entity_recognizers):
        self.assertEqual(self._result.entity_recognizers, expected_entity_recognizers)

    def test_validity_with_device_module_target(self):
        self._when_compile_service_interface(
            """
<service_interface>
  <validator name="RouteValidator">
    <parameters>
      <parameter predicate="navigation_origin"/>
      <parameter predicate="navigation_destination"/>
    </parameters>
    <target>
      <device_module device="NavigationDevice"/>
    </target>
  </validator>
</service_interface>
"""
        )
        self._then_service_interface_has_validities([
            ServiceValidatorInterface(
                "RouteValidator",
                target=DeviceModuleTarget("NavigationDevice"),
                parameters=[
                    ServiceParameter("navigation_origin", ParameterField.VALUE),
                    ServiceParameter("navigation_destination", ParameterField.VALUE),
                ]
            )
        ])

    def _then_service_interface_has_validities(self, expected_validities):
        self.assertEqual(self._result.validators, expected_validities)

    def test_frontend_target_for_action(self):
        self._when_compile_service_interface(
            """
<service_interface>
  <action name="CallContact">
    <parameters>
      <parameter predicate="contact_to_call"/>
    </parameters>
    <failure_reasons/>
    <target>
      <frontend/>
    </target>
  </action>
</service_interface>
"""
        )
        self._then_service_interface_has_actions([
            ServiceActionInterface(
                "CallContact",
                target=FrontendTarget(),
                parameters=[
                    ServiceParameter("contact_to_call"),
                ],
                failure_reasons=[]
            )
        ])

    def test_frontend_target_for_query(self):
        self._when_compile_service_interface_then_exception_is_raised_matching(
            """
<service_interface>
  <query name="available_contact">
    <parameters>
      <parameter predicate="contact_to_call"/>
    </parameters>
    <target>
      <frontend/>
    </target>
  </query>
</service_interface>
""", ViolatesSchemaException,
            "Expected service_interface.xml compliant with schema but it's in violation: Element 'frontend': "
            "This element is not expected. Expected is one of \( device_module, http \)., line 8"
        )

    def test_frontend_target_for_validator(self):
        self._when_compile_service_interface_then_exception_is_raised_matching(
            """
<service_interface>
  <validator name="RouteValidator">
    <parameters>
      <parameter predicate="navigation_origin"/>
      <parameter predicate="navigation_destination"/>
    </parameters>
    <target>
      <frontend/>
    </target>
  </validator>
</service_interface>
""", ViolatesSchemaException,
            "Expected service_interface.xml compliant with schema but it's in violation: Element 'frontend': "
            "This element is not expected. Expected is one of \( device_module, http \)., line 9"
        )

    def test_frontend_target_for_entity_recognizer(self):
        self._when_compile_service_interface_then_exception_is_raised_matching(
            """
<service_interface>
  <entity_recognizer name="ContactNameRecognizer">
    <target>
      <frontend/>
    </target>
  </entity_recognizer>
</service_interface>
""", ViolatesSchemaException,
            "Expected service_interface.xml compliant with schema but it's in violation: Element 'frontend': "
            "This element is not expected. Expected is one of \( device_module, http \)., line 5"
        )

    def test_http_target_for_action(self):
        self._when_compile_service_interface(
            """
<service_interface>
  <action name="CallContact">
    <parameters>
      <parameter predicate="contact_to_call"/>
    </parameters>
    <failure_reasons/>
    <target>
      <http endpoint="mock_endpoint"/>
    </target>
  </action>
</service_interface>
"""
        )
        self._then_service_interface_has_actions([
            ServiceActionInterface(
                "CallContact",
                target=HttpTarget("mock_endpoint"),
                parameters=[
                    ServiceParameter("contact_to_call"),
                ],
                failure_reasons=[]
            )
        ])

    def test_http_target_for_query(self):
        self._when_compile_service_interface(
            """
<service_interface>
  <query name="available_contact">
    <parameters>
      <parameter predicate="contact_to_call"/>
    </parameters>
    <target>
      <http endpoint="mock_endpoint"/>
    </target>
  </query>
</service_interface>
    """
        )
        self._then_service_interface_has_queries([
            ServiceQueryInterface(
                "available_contact",
                target=HttpTarget("mock_endpoint"),
                parameters=[
                    ServiceParameter("contact_to_call", ParameterField.VALUE),
                ]
            )
        ])

    def test_http_target_for_entity_recognizer(self):
        self._when_compile_service_interface(
            """
<service_interface>
  <entity_recognizer name="ContactNameRecognizer">
    <target>
      <http endpoint="mock_endpoint"/>
    </target>
  </entity_recognizer>
</service_interface>
    """
        )
        self._then_service_interface_has_entity_recognizers([
            ServiceEntityRecognizerInterface("ContactNameRecognizer", target=HttpTarget("mock_endpoint"))
        ])

    def test_http_target_for_validator(self):
        self._when_compile_service_interface(
            """
<service_interface>
  <validator name="RouteValidator">
    <parameters>
      <parameter predicate="navigation_origin"/>
      <parameter predicate="navigation_destination"/>
    </parameters>
    <target>
      <http endpoint="mock_endpoint"/>
    </target>
  </validator>
</service_interface>
"""
        )
        self._then_service_interface_has_validities([
            ServiceValidatorInterface(
                "RouteValidator",
                target=HttpTarget("mock_endpoint"),
                parameters=[
                    ServiceParameter("navigation_origin", ParameterField.VALUE),
                    ServiceParameter("navigation_destination", ParameterField.VALUE),
                ]
            )
        ])

    def test_play_audio_action_without_url_parameter(self):
        self._when_compile_service_interface_then_exception_is_raised_matching(
            """
<service_interface>
  <play_audio_action name="PlayAudio">
    <parameters/>
    <target>
      <frontend/>
    </target>
  </play_audio_action>
</service_interface>
""", ViolatesSchemaException,
            "Expected service_interface.xml compliant with schema but it's in violation: Element 'parameters': "
            "This element is not expected. Expected is \( audio_url_parameter \)., line 4"
        )

    def test_play_audio_action_with_parameters(self):
        self._when_compile_service_interface(
            """
<service_interface>
  <play_audio_action name="PlayAudio">
    <audio_url_parameter predicate="url_of_song"/>
    <parameters>
      <parameter predicate="artist"/>
    </parameters>
    <target>
      <frontend/>
    </target>
  </play_audio_action>
</service_interface>
"""
        )
        self._then_service_interface_has_actions([
            PlayAudioActionInterface(
                "PlayAudio",
                target=FrontendTarget(),
                parameters=[
                    ServiceParameter("artist"),
                ],
                audio_url_parameter=AudioURLServiceParameter("url_of_song")
            )
        ])

    def test_play_audio_action_and_custom_action_together(self):
        self._when_compile_service_interface(
            """
<service_interface>
  <play_audio_action name="PlayAudio">
    <audio_url_parameter predicate="url_of_song"/>
    <parameters>
      <parameter predicate="artist"/>
    </parameters>
    <target>
      <frontend/>
    </target>
  </play_audio_action>
  <action name="CallContact">
    <parameters>
      <parameter predicate="contact_to_call" format="grammar_entry"/>
    </parameters>
    <failure_reasons/>
    <target>
      <device_module device="CallDevice"/>
    </target>
  </action>
</service_interface>
"""
        )
        self._then_service_interface_has_actions([
            PlayAudioActionInterface(
                "PlayAudio",
                target=FrontendTarget(),
                parameters=[
                    ServiceParameter("artist"),
                ],
                audio_url_parameter=AudioURLServiceParameter("url_of_song")
            ),
            ServiceActionInterface(
                "CallContact",
                target=DeviceModuleTarget("CallDevice"),
                parameters=[
                    ServiceParameter("contact_to_call", ParameterField.GRAMMAR_ENTRY),
                ],
                failure_reasons=[]
            )
        ])

    def test_play_audio_action_with_invalid_target(self):
        self._when_compile_service_interface_then_exception_is_raised_matching(
            """
<service_interface>
  <play_audio_action name="PlayAudio">
    <audio_url_parameter predicate="url_of_song"/>
    <parameters>
      <parameter predicate="artist"/>
    </parameters>
    <target>
      <device_module device="AudioDevice"/>
    </target>
  </play_audio_action>
</service_interface>
""", ViolatesSchemaException,
            "Expected service_interface.xml compliant with schema but it's in violation: Element 'device_module': "
            "This element is not expected. Expected is \( frontend \)., line 9"
        )
